#ifndef MCMCPHASE_H
#define MCMCPHASE_H

#include "Phase.h"

class Model;
class InferenceTree;
class SequenceTable;



class MCMCphase : public Phase {

public:

    /** ***********************************************************************
     * randomTree
     * @input      table, the sequence table used to provide the labels
     * @input      maxBranchLengths, the biggest distance allowed between 2
     *             nodes
     * @input      outgroup, the outgroup used to standardize the tree before
     *             the method returns
     * @semantics  create a random tree, the internal branch are more likely
     *             to be shorter than the terminal ones
     ************************************************************************ */
    double runMCMC( InferenceTree * startingTree,
                    Model * startingModel, double initialBranchStep,
                    int burninCycles, int samplingCycles, int samplingPeriod,
                    const string& outgroup,
                    double branchLengthsUpperBound, string terminator );

    /** ***********************************************************************
     * randomTree
     * @input      table, the sequence table used to provide the labels
     * @input      maxBranchLengths, the biggest distance allowed between 2
     *             nodes
     * @input      outgroup, the outgroup used to standardize the tree before
     *             the method returns
     * @semantics  create a random tree, the internal branch are more likely
     *             to be shorter than the terminal ones
     ************************************************************************ */
    InferenceTree* randomTree( SequenceTable * table, double maxBranchLengths,
                               const string& outgroup );
                               
    /** ***********************************************************************
     * run
     * @input          argc, number of arguments in argv[]
     * @input          argv, an array of string arguments
     * @preconditions  argc is equal to 1 and argv[0] is the location
     *                 of a 'control-file' for a MCMC inference
     ************************************************************************ */
    void run( int argc, char * argv[] );
                               
};
#endif //MCMCPHASE_H




