#ifndef RNA7D_H
#define RNA7D_H

#include <math.h>

#include "RNA7A.h"

#include "matrixmath.h"

#include "SequenceTable.h"
#include "Util/array2D.h"



class RNA7D : public RNA7A {
private:
    /** .**********************************************************************
     * RNA7D
     * @input      registrationName, the name used to register to the model
     *             factory
     * @semantics  private constructor of the RNA 7D substitution
     *             model class, this construtor is used only once at launch
     *             time to register the prototype to the ModelFactory
     ************************************************************************ */
    RNA7D( const string & registrationName );

protected:
    /** .***********************************************************************
     * RNA7D
     * @input      a ParametersSet object to initialise the model
     * @pre        the set of parameters is consistent (see documentation)
     * @semantics  protected constructor of the RNA7D mutation model object :
     *             a model must be constructed via the model factory
     ************************************************************************ */
    RNA7D( ParametersSet & parameters );

public:

    /* .**********************************************************************
    * clone
    * @input      a ParametersSet object to initialise the new model
    * @pre        the set of parameters is consistent (see documentation)
    * @semantics  function used by the ModelFactory to clone the prototype and
    *             create a real model.
    ************************************************************************ */
    virtual Model * clone( ParametersSet & parameters ) const {
        return new RNA7D( parameters );
    }

    /** ************************************************************************
     * initialisation
     * @input      sequenceTable, the SequenceTable instance containing the
     *             species' sequences used to perform the inference
     *             (not compulsory)
     * @semantics  Initialise the model and its parameters according to
     *             the sequence table if provided.
     ************************************************************************ */
    virtual void initialisation( SequenceTable * sequenceTable = NULL, int modelId = 0 );

    /** ************************************************************************
     * ~RNA7D
     * @semantics  Destructor of the class Rna7a, free dynamic variables
     ************************************************************************ */
    virtual ~RNA7D();


    /** ************************************************************************
     * getName
     * @return  The name of the model
     ************************************************************************ */
    virtual string getName( void ) const;

protected:
    virtual void initMatrixIndex();

private:
    static RNA7D prototype;

};

#endif //RNA7D_H




