#ifndef INFERENCETREE_H
#define INFERENCETREE_H

#include <string>

#include "statlib.h"
#include "Node.h"

#include "Tree/Tree.h"


class Node;
class array_bag;
class randombox;

template < class Q >
class array2D;


template < class T >
class Array;


template < class Z >
class Array3D;
class SequenceTable;

class Model;


class InferenceTree : public Tree {

public:
    // Constructors


    InferenceTree();

    InferenceTree( string stree );

    InferenceTree( InferenceTree & );

    // Destructor
    ~InferenceTree();

    InferenceTree & operator = ( const InferenceTree & src );

    // Likelihood Calculation Methods
    vector < double > numeric_diffloglikelihood();
    vector < double > diffloglikelihood();
    vector < double > diffloglikelihood( vector < double > & );

    void loglikelihood_vector( vector < double > & lvector );
    array2D< double > hessian();
    double codifferential( int branch1, int branch2 );
    array3D < double > * getLeafCalculations();
    void loadLeafCalculations( array3D < double > * plikes );

    double variance();
    double variance( vector < double > & bl );
    double variance( InferenceTree & );

    /** ************************************************************************
     * normalize
     * @input   the id of the branch to normalize
     * @preconditions  the branch length is negative
     ************************************************************************ */
    int normalize( int branch, int left, int right );

    // Tree Query Methods

    bool swap_distance_two( int ibranch, bool up, int coin_flip, int rand1, int rand2 );

    // Branch Manipulation Methods
    bool loadLabels( SequenceTable & );

    // Returns the string representation of the tree

    // bool TopologyEqual(tree & tree2) ;

    // Clade Representatio building functions
    void build_clades( int, SequenceTable & );

    // Clade Query function
    bool clade_present( vector < int > &, string & );


    void reclaimSpace();
    void loadDataAndModel( SequenceTable * psequenceTable, Model * pmodel );
    void clade_list( array_bag & );

};

#endif //INFERENCETREE_H




