#ifndef TREE_H
#define TREE_H

using namespace std;

#include <string>
#include <vector>
#include <deque>


class Node;
class Model;
class SequenceTable;


class Tree {
protected:
    string outgroup;
    deque < Node * > nodeRefVector;
    Node * root;

public:
        
    deque < Node * > leafRefVector;

    Tree();

    Tree( string );

private:
    Tree(  const Tree & src );

public:
    ~Tree();

    double loglikelihood();

    double loglikelihood( vector < double > & bl );

    bool greater( string & ostring1, string & ostring2, string & soutg );

    bool standardizeTree( string newOutgroup );

    void addLeaf( const string & leafLabel, int branchId, double distParent = 0.1,
    double fracBranch = 0.5 );

    bool removeLeaf( const string & leafLabel );

    bool addClusteredNode( Node * node, int branchId, double distParent = 0.1,
    double fracBranch = 0.5 );

    Node * removeNode( const string & nodeLabel );

    bool is_internal_branch( int );

    bool swapBranches( int, int );

    bool isRooted();

    double insertBranch( int movedBranch, int destBranch, double insertionFrac );

    bool unroot();

    double getBranchLength( int branchId ) const;

    void setBranchLength( int branchId, double newLength );

    void getBranchVector( vector < double > & branchesLength ) const;

    void setBranchVector( const vector < double > & branchesLength );

    string toStringClade( vector < int > *, int );

    string toString( bool );

    string toStringTopology( bool );

    string toStringNumbered( bool );

    inline int getNumberNodes() const {
        return nodeRefVector.size();
    }

    inline int getNumberLeaves() const {
        return leafRefVector.size();
    }

    inline int getNumberBranches() const {
        return ( 2 * leafRefVector.size() - 3 ); // if unrooted tree
    }
    

    void createIndex();
    
protected:

    void retrieveBranchEnds( int branchId, int & fromNode, int & childNumber );

    bool installOutgroup( string newOutgroup );

public:
    void invalidateNodes();

    void invalidateLeaves();

    Model * t_model;
    SequenceTable * t_table;

};

#endif //TREE_H




