#include <algorithm>

/*
#include <dlfcn.h>
*/

#include <iostream>
#include <iomanip>
#include <ctype.h>
#include <functional>


#include "PatternDesign/Singleton.h"
#include "Util/randombox.h"

#include "Util/FileParser.h"
#include "Util/ParametersSet.h"

#include "Models/Model.h"
#include "Util/ModelFactory.h"

#include "Tree/InferenceTree.h"

using namespace std;


int main( int argc, char* argv[]){
  
    if ( argc != 2 ) {
        cerr << "usage : " << argv[0] << " control_file" << endl;
        exit( 1 );
    }

    // Create a FileParser from the control-file
    FileParser * fileParser = new FileParser( argv[1] );

    // Retrieve the control-file parameters
    ParametersSet parameters = fileParser->retrieveParametersSet();

    delete fileParser;
    
    string outgroup = parameters( "DATAFILE" ).stringParameter( "Outgroup" );
    
    /** ************************************************************************
     * Data File Initialization
     ************************************************************************ */

    // Read in molecular data and remove unknown characters from the data matrix
    string dataFile = parameters( "DATAFILE" ).stringParameter( "Data file" );
    SequenceTable * table = new SequenceTable( parameters( "DATAFILE" ) );

    /** ************************************************************************
     * Model Initialization
     ************************************************************************ */
    Singleton < ModelFactory > & modelFactory = Singleton < ModelFactory >::instance();
    Model* model = modelFactory.create( parameters( "MODEL" ).stringParameter( "Model" ),
         parameters( "MODEL" ) );
    model->initialisation( NULL );
    // Create a FileParser from the model parameters-file
    FileParser * modelFileParser = new FileParser( parameters.stringParameter("Model parameters file") );
    // Retrieve the model parameters
    ParametersSet modelParameters = modelFileParser->retrieveParametersSet();
    //initialise the model with them
    model->setModelParameters(modelParameters);

    /** ************************************************************************
     * Tree Initialization
     ************************************************************************ */
    string userTreeFileName = parameters.stringParameter( "Tree file" );
    ifstream userTreeFile( userTreeFileName.c_str() );
    string stringTree;
    FileParser::readTree(userTreeFile, stringTree);
    if ( stringTree.length() == 0 ){
        cerr << "Error while reading the tree in the file : "
             << userTreeFileName << endl;
        exit(1);
    }
    cout << stringTree << endl;
    InferenceTree* tree = new InferenceTree( stringTree );
    //tree->standardizeTree( outgroup );
    
    tree->loadDataAndModel(table, model);
    
    cout.setf(ios::fixed);
    cout << setprecision(4);
    cout << "likelihood = " << tree->loglikelihood() << endl;
}
