#include "Models/JC69.h"

#include <assert.h>

#include <iostream>
#include <iomanip>
#include <stdio.h>

#include "PatternDesign/Singleton.h"

#include "Util/ModelFactory.h"
#include "Models/Perturbator.h"

#include "Sequence/SequenceTable.h"

#include "Util/statlib.h"
#include "Util/randombox.h"
#include "Util/ParametersSet.h"


#define SQR(n) (n*n)

using namespace std;

//register the model to the model factory with the name JC69
JC69 JC69::prototype( "JC69" );

JC69::JC69( const string & registrationName ) : DnaModel( registrationName ){
//the private constructor is called by the prototype only
}

JC69::JC69( ParametersSet & parameters ) : DnaModel( parameters ){
//the normal constructor is called by the prototype's clone method
}

void JC69::initialisation( SequenceTable * sequenceTable, int modelId ) {
    //initialise the number of parameters
    numberRatesRatiosPerMatrix = 0;
    numberFrequencies = 0;
    //initialise the space for the parameters
    rateRatios = NULL;
    frequencies = NULL;
        
    //call the basic initialisation method
    DnaModel::initialisation( sequenceTable, modelId );
    
    //if a sequence table is provided initialise rate ratios and frequencies
    //accordingly
    if (sequenceTable){
        //estimation of all the rate ratios (and proportion of invariant sites
        //if relevant) from the empirical sequences
        vector < double > rate5;
        if ( invariantCategory ) {
            rate5 = retrieveEmpiricalRates( sequenceTable, modelId,
            & proportionInvariantSites );
        }
        else {
            //not useful
        }
    }
    //init the equivalency table (correspondance symbol/states)
    initEquivalencyTable();
    //first initialisation of the mutation rate matrix
    //(factor(s) used with the subtitution matrix to reach the specified
    //average substitution rate) and eigen system
    setAverageMutationRate( averageMutationRate );
}
  
JC69::~JC69() {
}

string JC69::getName( void ) const {
    char num[3];
    string name( "JC69" );
    if ( discreteGamma ) {
        sprintf( num, "%i", numberGammaCategories );
        name = name + string( " + dG" ) + string( num );
    }
    if ( invariantCategory == 1 )
        name = name + string( " + I" );
    return ( name );
}

void JC69::setAverageMutationRate( double newAverageRate, int symbolCategory ) {
    assert(symbolCategory==0);
    double mr;
    assert( newAverageRate > 0.0 );
    averageMutationRate = newAverageRate;

    // compute mutation rate for non-invariant sites
    mr = averageMutationRate / ( 1.0 - proportionInvariantSites );
    // if a gamma distribution is used, update the mutation rate
    // of all gamma categories
    if ( discreteGamma ) {
        // compute the gamma scale parameter beta
        double beta = alpha / mr;
        statlib::gamma_percentile_means( mutationRate, alpha, beta,
        numberGammaCategories );
        for ( int i = 0 ; i < numberGammaCategories; ++i ) {
            mutationRate[i] /= .750;
        }
    }
    else {
        mutationRate[0] = mr / .750;
    }
    setEigenMatrix();
}


void JC69::printParameters( ostream & outputStream ) const{
    MatrixModel::printParameters( outputStream );
    outputStream << "f[A] = .25    f[C] = .25    f[G] = .25    f[U] = .25    "
                 << endl;
    outputStream << "no rate parameter with the Jukes-Cantor model" << endl;
}



void JC69::setEigenMatrix() {
    for ( int category = 0; category < MAX( numberGammaCategories, 1 );
    category++ ) {
        rateMatrix[category] ( 0, 0 ) = mutationRate[category] * -.75;
        rateMatrix[category] ( 0, 1 ) = mutationRate[category] * .25;
        rateMatrix[category] ( 0, 2 ) = mutationRate[category] * .25;
        rateMatrix[category] ( 0, 3 ) = mutationRate[category] * .25;

        rateMatrix[category] ( 0, 0 ) = mutationRate[category] * .25;
        rateMatrix[category] ( 0, 1 ) = mutationRate[category] * -.75;
        rateMatrix[category] ( 0, 2 ) = mutationRate[category] * .25;
        rateMatrix[category] ( 0, 3 ) = mutationRate[category] * .25;

        rateMatrix[category] ( 0, 0 ) = mutationRate[category] * .25;
        rateMatrix[category] ( 0, 1 ) = mutationRate[category] * .25;
        rateMatrix[category] ( 0, 2 ) = mutationRate[category] * -.75;
        rateMatrix[category] ( 0, 3 ) = mutationRate[category] * .25;

        rateMatrix[category] ( 0, 0 ) = mutationRate[category] * .25;
        rateMatrix[category] ( 0, 1 ) = mutationRate[category] * .25;
        rateMatrix[category] ( 0, 2 ) = mutationRate[category] * .25;
        rateMatrix[category] ( 0, 3 ) = mutationRate[category] * -.75;
        
        for ( int i = 0; i < 4; ++i ) {
            eigenValues[category] [i] = 0.0;
            for ( int j = 0; j < 4; ++j ) {
                eigenMatrix[category] ( i, j ) = 0.0;
                ieigenMatrix[category] ( i, j ) = 0.0;
            }
        }
        eigenValues[category] [0] = 0.0;
        eigenValues[category] [1] = -mutationRate[category];
        eigenValues[category] [2] = -mutationRate[category];
        eigenValues[category] [3] = -mutationRate[category];


        // Update the eigenmatrix by hand
        eigenMatrix[category] ( 0, 0 ) = .25;
        eigenMatrix[category] ( 0, 1 ) = .25;
        eigenMatrix[category] ( 0, 2 ) = .25;
        eigenMatrix[category] ( 0, 3 ) = .25;

        eigenMatrix[category] ( 1, 0 ) = 0.0;
        eigenMatrix[category] ( 1, 1 ) = -1.0;
        eigenMatrix[category] ( 1, 2 ) = 0.0;
        eigenMatrix[category] ( 1, 3 ) = 1.0;

        eigenMatrix[category] ( 2, 0 ) = -1.0;
        eigenMatrix[category] ( 2, 1 ) = 0.0;
        eigenMatrix[category] ( 2, 2 ) = 1.0;
        eigenMatrix[category] ( 2, 3 ) = 0.0;

        eigenMatrix[category] ( 3, 0 ) = -.25;
        eigenMatrix[category] ( 3, 1 ) = .25;
        eigenMatrix[category] ( 3, 2 ) = -.25;
        eigenMatrix[category] ( 3, 3 ) = .25;

        // Inverse of the eigenmatrix
        ieigenMatrix[category] ( 0, 0 ) = 1.0;
        ieigenMatrix[category] ( 0, 1 ) = 0.0;
        ieigenMatrix[category] ( 0, 2 ) = -.5;
        ieigenMatrix[category] ( 0, 3 ) = -1.0;

        ieigenMatrix[category] ( 1, 0 ) = 1.0;
        ieigenMatrix[category] ( 1, 1 ) = -.5;
        ieigenMatrix[category] ( 1, 2 ) = 0.0;
        ieigenMatrix[category] ( 1, 3 ) = 1.0;

        ieigenMatrix[category] ( 2, 0 ) = 1.0;
        ieigenMatrix[category] ( 2, 1 ) = 0.0;
        ieigenMatrix[category] ( 2, 2 ) = .5;
        ieigenMatrix[category] ( 2, 3 ) = -1.0;

        ieigenMatrix[category] ( 3, 0 ) = 1.0;
        ieigenMatrix[category] ( 3, 1 ) = .5;
        ieigenMatrix[category] ( 3, 2 ) = 0.0;
        ieigenMatrix[category] ( 3, 3 ) = 1.0;
    } //for each category (except invariant)
}

Model * JC69::clone( ParametersSet & parameters ) const {
    if( parameters.find( "Number of frequencies categories" ) !=
            parameters.end() ){
        assert(parameters.intParameter("Number of frequencies categories") == 0);
    }
    else{
        parameters["Number of frequencies categories"] = "0";
    }
    if( parameters.find( "Number of rates matrices" ) !=
            parameters.end() ){
        assert(parameters.intParameter("Number of rates matrices") == 0);
    }
    else{
        parameters["Number of rates matrices"] = "0";
    }
    return new JC69( parameters );
}
