#include "Models/MixedPerturbator.h"

#include "Util/ParametersSet.h"

#include <iostream>

using namespace std;

MixedPerturbator::MixedPerturbator( int numberModels ):Perturbator(),
                                          averageRate(numberModels-1){
    lastModifiedModel = -1;
    models.clear();
    rndModels.clear();
    averageRateToModel.clear();
}

void MixedPerturbator::registerModel( Model* model, int modelPrio,
            int averageRatePrio, double averageRateStep,
            double minAverageRatesAcceptanceRate, double maxAverageRatesAcceptanceTol){
    if (models.size()){
        char label[40];
        sprintf( label, "Model%d/model1 average mutation rate ratio",
                        models.size()+1 );
        averageRate[models.size()-1] = model->getAverageMutationRate(0);
        registerGaussParameter( label, &(averageRate[models.size()-1]),
                          averageRatePrio, averageRateStep, 0.0, 100.0,
                          minAverageRatesAcceptanceRate, maxAverageRatesAcceptanceTol );
        //store the model corresponding to each averageRate in rndTable
        //model 0 = average rate 1.0
        averageRateToModel.insert(averageRateToModel.end(), averageRatePrio, models.size());
        assert( averageRateToModel.size() == rndTable.size() );
    }
    models.push_back( model );
    rndModels.insert(rndModels.end(), modelPrio, model );
}
        
double MixedPerturbator::perturb(){
    if( rndTable.size() || rndModels.size() ){
        assert(lastModified == -1);
        assert(lastModifiedModel == -1);
        int nb = (int)(rand.ran()*(double)(rndTable.size()+rndModels.size()));
        if ( nb >= (int)rndTable.size() ){
            lastModifiedModel = nb - (int)rndTable.size();
            return rndModels[lastModifiedModel]->perturb();
        }
        else{
            lastModified = nb;
            rndTable[lastModified]->perturb();
            models[averageRateToModel[lastModified]]->
               setAverageMutationRate(averageRate[averageRateToModel[lastModified]-1],0);
            return 0.0;
        }
    }
    return 0.0;
}

bool MixedPerturbator::validatePerturbation(bool validation){
    if( rndTable.size() || rndModels.size() ){
        if (lastModified != -1){
            assert(lastModifiedModel==-1);
            bool res = rndTable[lastModified]->validatePerturbation(validation);
            if (!validation){
                models[averageRateToModel[lastModified]]->
                   setAverageMutationRate(
                        averageRate[averageRateToModel[lastModified]-1],0);
            }
            lastModified = -1;
            return res;
        }
        else{
            bool res = rndModels[lastModifiedModel]->validatePerturbation(validation);
            lastModifiedModel = -1;
            return res;
        }
    }
    return true;
}

void MixedPerturbator::printParameters(ostream& outputStream) {
    Perturbator::printParameters( outputStream );
    for ( unsigned int i = 0; i < models.size(); ++i ){
        outputStream << "MODEL " << i+1 << endl;
        models[i]->printPerturbationParameters( outputStream );
    }
}

