#include "Models/Perturbator.h"

#include "Util/ParametersSet.h"

#include "Models/PerturbatorParameter.h"

#include "Models/PerturbatorFrequenciesParameter.h"
#include "Models/PerturbatorGaussParameter.h"


#include <iostream>

using namespace std;

Perturbator::Perturbator():
    rand(Singleton < randombox >::instance()){
    lastModified = -1;
    parameters.clear();
    rndTable.clear();
}

void Perturbator::registerFrequencies( string name, vector<double> * frequencies,
    int freqPrio, double freqStep,
    double minFreqAcceptanceRate, double maxFreqAcceptanceRate ){

    PerturbatorFrequenciesParameter* freqParam = new
            PerturbatorFrequenciesParameter( name, frequencies,
                freqStep, minFreqAcceptanceRate, maxFreqAcceptanceRate );
    parameters.insert(parameters.end(), freqParam);
    rndTable.insert( rndTable.end(), freqPrio, freqParam );
}

void Perturbator::registerGaussParameter( string name, double* element,
    int elementPrio, double elementStep, double min, double max,
    double minElementAcceptanceRate, double maxElementAcceptanceRate ){
    PerturbatorGaussParameter* gaussParam = new
            PerturbatorGaussParameter( name, element, elementStep, min, max,
                minElementAcceptanceRate, maxElementAcceptanceRate );
    parameters.insert(parameters.end(), gaussParam);
    rndTable.insert( rndTable.end(), elementPrio, gaussParam );    
}

double Perturbator::perturb(){
    if (rndTable.size()){
        assert(lastModified == -1);
        lastModified = (int)(rand.ran()*(double)rndTable.size());
        return rndTable[lastModified]->perturb();
    }
    return 0.0;
}

bool Perturbator::validatePerturbation(bool validation){
    if (rndTable.size()){
        assert(lastModified != -1);
        bool res = rndTable[lastModified]->validatePerturbation( validation );
        lastModified = -1;
        return res;
    }
    return true;
}

void Perturbator::stopBurn(){
    for ( vector<PerturbatorParameter*>::iterator iter = parameters.begin();
          iter != parameters.end(); ++iter ){
        (*iter)->stopBurn();
    }
}

void Perturbator::printParameters(ostream& outputStream){
    for ( vector<PerturbatorParameter*>::iterator iter = parameters.begin();
          iter != parameters.end(); ++iter ){
        (*iter)->printParameter( outputStream );
    }
}
