#include "Tree/InferenceTree.h"

#include <math.h>
#include <assert.h>
#include <time.h>

using namespace std;

#include <string>

#include <iostream>

#include "Util/statlib.h"

#include "Util/array2D.h"
#include "Util/array_bag.h"

#include "Util/randombox.h"
#include "PatternDesign/Singleton.h"

#include "Tree/Node.h"

#include "Models/Model.h"

class SequenceTable;


// Creates an empty tree
InferenceTree::InferenceTree() : Tree() {
}


// Creates a tree from its string representation
InferenceTree::InferenceTree( string stree ) : Tree( stree ) {
}


// Constructs a tree by copying another tree
InferenceTree::InferenceTree( InferenceTree & src ) : Tree() {

    outgroup = src.outgroup;
    // since nodes have a reference to the tree the copy will be different
    root = new Node( this );
    if ( src.t_table ) {
        assert( src.t_model );
        // same model and sequence table (can be NULL)
        loadDataAndModel( src.t_table, src.t_model );
    }
    root->recursiveCopy( src.root, NULL );
    createIndex();
}

array3D < double > * InferenceTree::getLeafCalculations() {
    assert( 0 );

    /*
    array3D < double > * leaflikelihoods = new array3D < double > [nleaves];
    int i;
    for ( i = 0; i < nleaves; i++ )
    leaflikelihoods[i] = leaf[leaf[i]->cnumber - 1]->plikelihood;
    return leaflikelihoods;
    */

    return NULL;
}

void InferenceTree::loadLeafCalculations( array3D < double > * plikes ) {
    assert( 0 );
    plikes = plikes;

    /*
    int i;
    for ( i = 0; i < nleaves; i++ ) {
    leaf[i]->plikelihood = plikes[leaf[i]->cnumber - 1];
    leaf[i]->calculated = true;
    }
    return;
    */
}

InferenceTree & InferenceTree::operator = ( const InferenceTree & src ) {

    // No self assignment
    if ( root == src.root )
        return ( * this );

    // Special case (Efficient rearrangement of the leaves)
    // since the partialLikelihood vector is invariant
    if ( ( src.t_model == t_model ) && ( src.t_table == t_table ) &&
    ( t_model ) && ( t_table ) ) {
        assert( leafRefVector.size() == src.leafRefVector.size() );
        for ( unsigned int i = 0; i < leafRefVector.size(); ++i ) {
            bool found = false;
            unsigned int j = i;
            while ( ( !found ) && ( j < leafRefVector.size() ) ) {
                if ( leafRefVector[j]->label ==
                src.leafRefVector[i]->label ) {
                    found = true;
                    if ( j != i ) { // swap
                        Node * temp = leafRefVector[i];
                        leafRefVector[i] = leafRefVector[j];
                        leafRefVector[j] = temp;
                    }
                }
                ++j;
            } // end while
            assert( found );
            leafRefVector[i]->index = i;
            if (!leafRefVector[i]->calculated &&
                    src.leafRefVector[i]->calculated) {
                leafRefVector[i]->partialLikelihood =
                        src.leafRefVector[i]->partialLikelihood;
                leafRefVector[i]->calculated = true;
            }
            assert( leafRefVector[i]->identical ==
                       src.leafRefVector[i]->identical );
        }

        for ( unsigned int i = 0; i < nodeRefVector.size(); ++i ) {
            nodeRefVector[i]->label = src.nodeRefVector[i]->label;
            for ( int childId = 0; childId < 3; ++childId ) {
                if ( src.nodeRefVector[i]->child[childId] == NULL ) {
                    nodeRefVector[i]->child[childId] = NULL;
                }
                else {
                    nodeRefVector[i]->child[childId] =
                    ( src.nodeRefVector[i]->child[childId]->isLeaf() ?
                    leafRefVector[src.nodeRefVector[i]->child[childId]->index] :
                    nodeRefVector[src.nodeRefVector[i]->child[childId]->index] );
                    nodeRefVector[i]->child[childId]->parent = nodeRefVector[i];
                    nodeRefVector[i]->child[childId]->pdistance =
                    src.nodeRefVector[i]->cdistance[childId];
                    nodeRefVector[i]->cdistance[childId] =
                    src.nodeRefVector[i]->cdistance[childId];
                }
            }
            nodeRefVector[i]->identical = src.nodeRefVector[i]->identical;

            if ( src.nodeRefVector[i]->calculated ) {
                nodeRefVector[i]->partialLikelihood =
                    src.nodeRefVector[i]->partialLikelihood;
            }
            nodeRefVector[i]->calculated = src.nodeRefVector[i]->calculated;
        }

        root = nodeRefVector[nodeRefVector.size() - 1];
        return ( * this );
    }
    //general case
    assert( root == NULL );
    if ( root == NULL ) {
        root = new Node( this );
        if ( src.t_table ) {
            assert( src.t_model );
            // same model and sequence table (can be NULL)
            loadDataAndModel( src.t_table, src.t_model );
        }
        root->recursiveCopy( src.root, NULL );
        createIndex();
    }
    return ( * this );
}

// Destructor
InferenceTree::~InferenceTree() {
}

// Clade
void InferenceTree::build_clades( int mode, SequenceTable & data ) {
    assert( 0 );
    mode = mode;
    data = data;

    /*
    int i;
    for ( i = 0; i < nleaves; i++ ) {
    leaf[i]->label = leaf[i]->toStringTopology();
    leaf[i]->build_clade( mode, data );
    }
    for ( i = 0; i < nnodes; i++ ) {
    node[i]->label = node[i]->toStringTopology();
    node[i]->build_clade( mode, data );
    }
    return;
    */
}

void InferenceTree::clade_list( array_bag & arr ) {
    assert( 0 );
    arr = arr;
    //    int i;
    //    for ( i = nnodes - 1; i >= 0; i-- ) arr.add( node[i]->clade );
    //    return;
}


bool InferenceTree::clade_present( vector < int > & clade_rep,
string & sclade_rep ) {
    assert( 0 );
    clade_rep = clade_rep;
    sclade_rep = sclade_rep;

    /*
    int i = 0, isize;
    int nsize, ssize = size_of_array( clade_rep );
    sclade_rep = string( "" );
    Node * cnode = root;


    while ( i < 4 ) {
    if ( cnode->child[i] != NULL ) {
    isize = intersect_set_size( cnode->child[i]->clade, clade_rep );
    nsize = size_of_array( cnode->child[i]->clade );
    if ( ( isize == nsize ) && ( isize == ssize ) ) {
    sclade_rep = cnode->child[i]->label;
    return true;
    }
    if ( ( isize < nsize ) && ( isize == ssize ) ) {
    cnode = cnode->child[i];
    i = 0;
    }
    else {
    if ( ( isize < ssize ) && ( isize > 0 ) ) return false;
    else
    i++;
    }
    }
    else {
    i++;
    }
    } // while (i<4)
    */

    return false;
}

// Invalidate internal node calculations

// Invalidate leaf node calculations

void InferenceTree::reclaimSpace() {
    unsigned int i;
    assert( 0 );
    for ( i = 0; i < leafRefVector.size(); i++ ) {
        //    leafRefVector[i]->prior.reclaimSpace();
        //    leafRefVector[i]->belief.reclaimSpace();
    }
    for ( i = 0; i < nodeRefVector.size(); i++ ) {
        //    nodeRefVector[i]->prior.reclaimSpace();
        //    nodeRefVector[i]->belief.reclaimSpace();
    }
}

// Protected Methods (Unrooted_tree Class)

bool InferenceTree::loadLabels( SequenceTable & table ) {
    assert( 0 );
    table = table;

    /*
    int i, j;
    char charnum[26];
    for ( i = 0; i < nleaves; i++ ) {
    for ( j = 0; j < ( int )leaf[i]->label.length(); j++ )
    charnum[j] = leaf[i]->label[j];
    charnum[j] = 0;
    leaf[i]->label = table.species[atoi( charnum ) - 1];

    }
    */

    return true;
}

// returns branch lengths of the tree in the vector bl



void InferenceTree::loadDataAndModel( SequenceTable * psequenceTable,
Model * pmodel ) {
    assert ( t_table == NULL );
    assert ( t_model == NULL );
    t_table = psequenceTable;
    t_model = pmodel;
    
    invalidateNodes();
    invalidateLeaves();
    // WARNING, the order to compute the identical is IMPORTANT
    for ( unsigned int i = 0; i < leafRefVector.size(); ++i ) {
        leafRefVector[i]->initialisation( true );
    }
    for ( unsigned int i = 0; i < nodeRefVector.size(); ++i ) {
        nodeRefVector[i]->initialisation( false );
        nodeRefVector[i]->computeIdentical();
    }
}


/*
void tree::loadModel( Model & newModel ) {
unsigned int i;
assert (t_model==NULL);
invalidateNodes();
invalidateLeaves();
if (t_table){
for ( i = 0; i < t_table->getSequencesLength(); ++i ) {
leafRefVector[i]->identical[i] = t_model->getSymbolNumber( sequence[i] );
}

}

*/

double InferenceTree::variance( vector < double > & branchLengths ) {

    vector < double > oldBranchLengths;
    double ans;

    getBranchVector( oldBranchLengths );
    setBranchVector( branchLengths );
    ans = variance();
    setBranchVector( oldBranchLengths );

    return ( ans );
}

double InferenceTree::variance() {
    assert(0);
    /*
    vector < double > likelihoods, rate_prob;
    int i, nucleotide, nstates, slength, nrates, rate;
    int number_of_invariant_bases;
    double sum, site, mean, sumv;
    if ( t_model == NULL ) {
    cerr << "Mutation model not loaded " << endl;
    exit( 1 );
    }
    if ( t_table == NULL ) {
    cerr << "Sequence data not loaded " << endl;
    exit( 1 );
    }

    number_of_invariant_bases = t_table->getNumberInvariantBases();
    nrates = t_model->getNumberRateCategories();
    nstates = t_model->getNumberStates();
    slength = t_table->getSequencesLength();
    likelihoods.resize( slength );
    rate_prob.resize( nrates );


    for ( i = 0; i < nrates; i++ )
    rate_prob[i] = t_model->getRateCategoryProbability( i );

    mean = loglikelihood() / t_table->getRawLength();
    //cerr << "Mean = " << mean << endl ;
    sumv = 0;
    for ( i = 0; i < ( slength - number_of_invariant_bases ); i++ ) {
    sum = 0.0;
    for ( rate = 0; rate < nrates; rate++ ) {
    site = 0.0;
    for ( nucleotide = 0; nucleotide < nstates; nucleotide++ )
    site += ( t_model->getFrequency( nucleotide, rate ) *
    root->plikelihood( rate, i, nucleotide ) );
    sum += ( rate_prob[rate] * site );
    }
    sumv += ( ( log( sum ) - mean ) * ( log( sum ) - mean ) );
    }

    for ( ; i < slength; i++ ) {
    sum = 0.0;
    for ( rate = 0; rate < nrates; rate++ ) {
    site = 0.0;
    for ( nucleotide = 0; nucleotide < nstates; nucleotide++ )
    site += ( t_model->getFrequency( nucleotide, rate ) *
    root->plikelihood( rate, i, nucleotide ) );
    sum += ( rate_prob[rate] * site );
    }
    sumv += ( ( ( log( sum ) - mean ) * ( log( sum ) - mean ) ) *
    t_table->getNumberConstantSites( leaf[0]->sequence[i] ) );
    }

    return ( ( ( t_table->getRawLength() ) / ( t_table->getRawLength() - 1 ) )
    * sumv );
    */

    return 0;
}


double InferenceTree::variance( InferenceTree & tree2 ) {
    assert(0);
    tree2 = tree2;
    
/*
    if ( t_model == NULL ) {
        cerr << "Mutation model not loaded into tree 1" << endl;
        exit( 1 );
    }
    if ( t_table == NULL ) {
        cerr << "Sequence data not loaded into tree 1" << endl;
        exit( 1 );
    }
    if ( tree2.t_model == NULL ) {
        cerr << "Mutation model not loaded into tree 2" << endl;
        exit( 1 );
    }
    if ( tree2.t_table == NULL ) {
        cerr << "Sequence data not loaded into tree 2" << endl;
        exit( 1 );
    }

    if ( t_model->getName() != tree2.t_model->getName() ) {
        cerr << "Tree 1 and Tree 2 have incompatible models" << endl;
        exit( 1 );
    }

    double * rateProb[2][numberSymbolCategory];
    double * * frequency[2][numberSymbolCategory];
    // numberInvariantBases is the number of symbols found invariant at least
    // in one site
    int numberInvariantBases = t_table->getNumberInvariantBases();

    int sequenceLength = t_table->getSequencesLength();

    numberInvariantBases = t_table->getNumberInvariantBases();
    nrates = t_model->getNumberRateCategories();
    nstates = t_model->getNumberStates();
    slength = t_table->getSequencesLength();
    likelihood1.resize( slength );
    likelihood2.resize( slength );
    rate_prob1.resize( nrates );
    rate_prob2.resize( nrates );

    mean = 0;
    for ( i = 0; i < nrates; i++ ) {
    rate_prob1[i] = t_model->getRateCategoryProbability( i );
    rate_prob2[i] = tree2.t_model->getRateCategoryProbability( i );
    }

    for ( i = 0; i < ( slength - number_of_invariant_bases ); i++ ) {
    likelihood1[i] = likelihood2[i] = 0.0;
    for ( rate = 0; rate < nrates; rate++ ) {
    site1 = site2 = 0.0;
    for ( nucleotide = 0; nucleotide < nstates; nucleotide++ ) {
    site1 += ( t_model->getFrequency( nucleotide, rate ) *
    root->plikelihood( rate, i, nucleotide ) );
    site2 += ( tree2.t_model->getFrequency( nucleotide, rate ) *
    tree2.root->plikelihood( rate, i, nucleotide ) );
    }
    likelihood1[i] += ( rate_prob1[rate] * site1 );
    likelihood2[i] += ( rate_prob2[rate] * site2 );
    }
    mean += ( log( likelihood1[i] ) - log( likelihood2[i] ) );
    }
    for ( ; i < slength; i++ ) {
    likelihood1[i] = likelihood2[i] = 0.0;
    for ( rate = 0; rate < nrates; rate++ ) {
    site1 = site2 = 0.0;
    for ( nucleotide = 0; nucleotide < nstates; nucleotide++ ) {
    site1 += ( t_model->getFrequency( nucleotide, rate ) *
    root->plikelihood( rate, i, nucleotide ) );
    site2 += ( tree2.t_model->getFrequency( nucleotide, rate ) *
    tree2.root->plikelihood( rate, i, nucleotide ) );
    }
    likelihood1[i] += ( rate_prob1[rate] * site1 );
    likelihood2[i] += ( rate_prob2[rate] * site2 );
    }
    mean += ( ( log( likelihood1[i] ) - log( likelihood2[i] ) ) *
    t_table->getNumberConstantSites( leaf[0]->sequence[i] ) );
    }
    mean /= t_table->getRawLength();
    //cerr << "Mean = " << mean << endl ;
    sumv = 0.0;
    for ( i = 0; i < ( slength - number_of_invariant_bases ); i++ )
    sumv += pow( ( log( likelihood1[i] ) - log( likelihood2[i] ) -
    mean ), 2.0 );
    for ( ; i < slength; i++ )
    sumv += pow( ( log( likelihood1[i] ) - log( likelihood2[i] ) -
    mean ), 2.0 ) *
    t_table->getNumberConstantSites( leaf[0]->sequence[i] );
    // cerr << "Sumv = " << sumv << endl ;
    return ( ( t_table->getRawLength() / ( t_table->getRawLength() - 1 )
    ) * sumv );
    */

    return 0;
}


array2D< double > InferenceTree::hessian() {
    assert( 0 );
    array2D < double > ret;
    return ret;

    /*
  int num_branches = number_of_branches() ;
  Array2D<double> ret(num_branches , num_branches)  ;
  vector <double> lvector , darray[num_branches] , coarray[num_branches][num_branches];
  int i , j , site , nstates = t_model->number_of_states() , nucleotide;
  int nrates = t_model->number_of_rate_categories() , rate ;
  double temp , tempn , tempr , likelihood , sum;
  double frequency[nstates] ; 
  double rate_prob[nrates] ;
  int number_of_invariant_bases = t_table->getNumberInvariantBases() ;
  int slength = t_table->mod_sequence_length ;
  
  likelihood_array(larray) ;
  for (i=0 ; i<num_branches ; i++)
    differential_likelihood_array(darray[i],i) ;
  
  for (i=0 ; i<num_branches ; i++)
    for (j=0 ; j<num_branches ; j++) {
      //invalidate_nodes() ;
      cogerential_array(coarray[i][j],i,j) ;
      //if (i!=j)
      //	coarray[j][i] = coarray[i][j] ;
    }
     
    
  
  for (i=0 ; i<num_branches ; i++)
    for (j=0 ; j<num_branches ; j++) {
      sum = 0.0 ;
      for (site=0 ; site<(slength - number_of_invariant_bases) ; site++)
	sum +=  ((1.0/larray[site]) * coarray[i][j][site]) - (pow(larray[site],-2.0)*darray[i][site]*darray[j][site]) ;            
      for (site=(slength - number_of_invariant_bases) ; site<slength ; site++)
      	sum += (((1.0/larray[site]) * coarray[i][j][site]) - (pow(larray[site],-2.0)*darray[i][site]*darray[j][site])) * t_table->getNumberConstantSites(leaf[0]->sequence[site]);      
      ret(i,j) = sum ;
    
    }   
    
  for (i=0 ; i<num_branches ; i++)
    for (j=0 ; j<num_branches ; j++)
      ret(i,j) = ret(i,j) * -1.0 ;
  //  ret = inverse(ret) ;
  return ret ;
    */

    
}

double InferenceTree::codifferential( int branch1, int branch2 ) {
    assert( 0 );
    branch1 = branch1;
    branch2 = branch2;

    /*
    int node1, node2, nodebranch1, nodebranch2;
    int num_branches = 2 * nleaves - 3;
    int i, site, nstates = t_model->getNumberStates(), nucleotide;
    int nrates = t_model->getNumberRateCategories(), rate;
    double temp, tempn, tempr, likelihood;
    double rate_prob[nrates];
    int slength = t_table->getSequencesLength();
    int number_of_invariant_bases = t_table->getNumberInvariantBases();

    Node * tempnode;
    if ( branch1 > num_branches - 3 ) {
    node1 = nnodes - 1;
    nodebranch1 = branch1 - ( num_branches - 3 );
    }
    else {
    node1 = branch1 / 2;
    nodebranch1 = branch1 % 2;
    }
    if ( branch2 > num_branches - 3 ) {
    node2 = nnodes - 1;
    nodebranch2 = branch2 - ( num_branches - 3 );
    }
    else {
    node2 = branch2 / 2;
    nodebranch2 = branch2 % 2;
    }

    tempnode = node[node1];
    while ( tempnode != NULL ) {
    tempnode->calculated = false;
    tempnode = tempnode->parent;
    }
    tempnode = node[node2];
    while ( tempnode != NULL ) {
    tempnode->calculated = false;
    tempnode = tempnode->parent;
    }
    for ( i = 0; i < nleaves; i++ ) {
    if ( !( leaf[i]->calculated ) ) leaf[i]->likelihood();
    }
    for ( i = 0; i < nnodes; i++ ) {
    if ( ( i != node1 ) && ( i != node2 ) ) {
    if ( !( node[i]->calculated ) ) node[i]->likelihood();
    }
    else {
    if ( node1 == node2 ) {
    if ( nodebranch1 == nodebranch2 )
    node[i]->diff2likelihood( nodebranch1 );
    else
    node[i]->codiff_likelihood( nodebranch1, nodebranch2 );
    }
    else {
    if ( i == node1 ) node[i]->difflikelihood( nodebranch1 );
    else
    node[i]->difflikelihood( nodebranch2 );
    }
    }
    }

    temp = tempn = likelihood = 0.0;

    for ( rate = 0; rate < nrates; rate++ ) {
    rate_prob[rate] = t_model->getRateCategoryProbability( rate );
    }


    for ( site = 0; site < ( slength -
    number_of_invariant_bases ); site++ ) {
    tempr = 0.0;
    for ( rate = 0; rate < nrates; rate++ ) {
    tempn = 0.0;
    for ( nucleotide = 0; nucleotide < nstates; nucleotide++ )
    tempn += ( t_model->getFrequency( nucleotide, rate ) *
    root->plikelihood( rate, site, nucleotide ) );
    tempr += ( rate_prob[rate] * tempn );
    } // end rate

    likelihood += log( tempr );
    }

    for ( site = ( slength - number_of_invariant_bases ); site <
    slength; site++ ) {
    tempr = 0.0;
    for ( rate = 0; rate < nrates; rate++ ) {
    tempn = 0.0;
    for ( nucleotide = 0; nucleotide < nstates; nucleotide++ )
    tempn += ( t_model->getFrequency( nucleotide, rate ) *
    root->plikelihood( rate, site, nucleotide ) );
    tempr += rate_prob[rate] * tempn;
    } // end rate

    likelihood += ( log( tempr ) *
    t_table->getNumberConstantSites( leaf[0]->sequence[site] ) );
    }

    return ( likelihood );
    */

    return 0.0;
}

void InferenceTree::loglikelihood_vector( vector < double > & lvector ) {
    assert( 0 );
    lvector = lvector;

    /*
    if ( t_model == NULL ) {
    cerr << "Mutation model not loaded " << endl;
    exit( 1 );
    }
    if ( t_table == NULL ) {
    cerr << "Sequence data not loaded " << endl;
    exit( 1 );
    }

    int i, site, nstates = t_model->getNumberStates(), nucleotide;
    int nrates = t_model->getNumberRateCategories(), rate;
    double temp, tempn, tempr, likelihood;
    double rate_prob[nrates];
    int slength = t_table->getSequencesLength();


    for ( i = 0; i < nleaves; i++ ) {
    if ( !( leaf[i]->calculated ) ) {
    leaf[i]->likelihood();
    }
    }
    for ( i = 0; i < nnodes; i++ ) {
    if ( !( node[i]->calculated ) ) {
    node[i]->likelihood();
    }
    }

    temp = tempn = likelihood = 0.0;

    for ( rate = 0; rate < nrates; rate++ ) {
    rate_prob[rate] = t_model->getRateCategoryProbability( rate );
    }
    lvector.resize( slength );

    for ( site = 0; site < slength; site++ ) {
    lvector[site] = 0.0;
    tempr = 0.0;
    for ( rate = 0; rate < nrates; rate++ ) {
    tempn = 0.0;
    for ( nucleotide = 0; nucleotide < nstates; nucleotide++ )
    tempn += ( t_model->getFrequency( nucleotide, rate ) *
    root->plikelihood( rate, site, nucleotide ) );
    tempr += ( rate_prob[rate] * tempn );
    } // end rate

    lvector[site] = tempr;
    }
    */
}

// Likelihood Calculation Methods


vector < double > InferenceTree::numeric_diffloglikelihood() {

    assert( 0 );

    vector < double > ans;

    /*
    int i;
    double step = 1e-8, old, olikelihood, nlikelihood1, nlikelihood2;

    ans.resize( number_of_branches() );

    olikelihood = loglikelihood();
    for ( i = 0; i < number_of_branches(); i++ ) {
    old = get_branch( i );
    set_branch( i, old + step );
    nlikelihood2 = loglikelihood();
    set_branch( i, old - step );
    nlikelihood1 = loglikelihood();

    ans[i] = ( nlikelihood2 - nlikelihood1 ) / ( 2.0 * step );
    set_branch( i, old );
    }

    */

    return ( ans );
}

vector < double > InferenceTree::diffloglikelihood( vector < double > &
branchLengths ) {
    vector < double > oldBranchLengths;
    vector < double > partialGradients;

    getBranchVector( oldBranchLengths );
    setBranchVector( branchLengths );
    partialGradients = diffloglikelihood();
    setBranchVector( oldBranchLengths );

    return ( partialGradients );
}



vector < double > InferenceTree::diffloglikelihood() {
    assert ( t_model && t_table != NULL );
    vector < double > gradientVector( getNumberBranches() );
    for ( int i = 0; i < getNumberBranches(); ++i ){
        gradientVector[i] = 0.0;
    }
    

    //for each model
    for ( unsigned int modelId = 0; modelId < t_table->getNumberModels(); ++modelId ){

        unsigned int sequencesLength = t_table->getSequencesLength(modelId);
        unsigned int invariantsLength = t_table->getInvariantsLength(modelId);
        unsigned int totalLength = sequencesLength + invariantsLength;
        int numberLeaves = getNumberLeaves();
        int numberInternalNodes = getNumberNodes();
        
        int numberStates = t_model->getNumberStates(modelId);
        int numberRatesCategories = t_model->getNumberRateCategories(modelId);
        array2D <double> priorLeaves[numberLeaves][totalLength];
        array2D <double> priorNodes[numberInternalNodes][totalLength];
        array2D <double> beliefNodes[numberInternalNodes][totalLength];
    

        //BEWARE, the order of processing is important
        //leafRefVector and nodeRefVector are made in order to
        //compute the children likelihood before
        for ( unsigned int i = 0; i < leafRefVector.size(); ++i ) {
            if ( !( leafRefVector[i]->calculated ) ) {
                leafRefVector[i]->likelihood();
            }
            for ( unsigned int j = 0; j< totalLength; ++j ){
                priorLeaves[i][j].resize( numberRatesCategories,
                                           numberStates );
            }
        }
        for ( unsigned int i = 0; i < nodeRefVector.size(); ++i ) {
            if ( !( nodeRefVector[i]->calculated ) ) {
                nodeRefVector[i]->likelihood();
            }
            for ( unsigned int j = 0; j< totalLength; ++j ){
                priorNodes[i][j].resize( numberRatesCategories,
                                          numberStates );
                beliefNodes[i][j].resize( numberRatesCategories,
                                           numberStates );
            }
        }

        // Calculate the beliefs at the internal nodes
        for ( int rate = 0; rate < numberRatesCategories; ++rate ) {
            // for the root
            for ( unsigned int site = 0; site < totalLength; ++site ) {
                for ( int state = 0; state < numberStates; ++state ) {
                   beliefNodes[getNumberNodes()-1][site]( rate, state ) =
                        t_model->getFrequency( state, rate, modelId) *
                            root->partialLikelihood[modelId](site, state, rate );            
#ifdef DEBUG
                    assert( !isnan( 
                      beliefNodes[getNumberNodes()-1][site]( rate, state ) ) ) ;
                    assert( !isinf(
                      beliefNodes[getNumberNodes()-1][site]( rate, state ) ) ) ;
#endif
                }
            } // end root

        
            // for other internal nodes
            for ( int i = getNumberNodes() - 2; i >= 0; --i ) {
                // Find Parent index and which branch this node lies on
                int parentIndex = nodeRefVector[i]->parent->index; 
                int branchId = -1;
                for ( int j = 0; j < 3; j++ ) { 
                    if ( nodeRefVector[parentIndex]->child[j] ==
                         nodeRefVector[i] ){
                         branchId = j;
                         j = 3;
                    }
                }
#ifdef DEBUG
                assert( branchId != -1 );
#endif
                // branch length
                double branchLength =
                    nodeRefVector[parentIndex]->cdistance[branchId];
        
                // Initialize lookup table
                array3D < double > lookup( numberRatesCategories, numberStates,
                                           numberStates );
                for ( int initialState = 0; initialState < numberStates;
                      ++initialState ) {
                    for ( int finalState = 0; finalState < numberStates;
                          ++finalState ) {
                        lookup( rate, initialState, finalState ) =
                        t_model->probability( initialState, finalState,
                                              branchLength, rate, modelId );
                    }
                }        
        
                for ( unsigned int site = 0; site < totalLength; ++site ){
                    double temp[numberStates];
                    for ( int initialState = 0; initialState < numberStates;
                          ++initialState ){
                        double contrib = 0.0;
                        for ( int finalState = 0; finalState < numberStates;
                              ++finalState ){
                            contrib +=
                                lookup(rate, initialState, finalState ) *
                                nodeRefVector[i]->
                                partialLikelihood[modelId]( site, finalState, rate);
                        }
                        if (beliefNodes[parentIndex][site](rate,initialState)==
                            0.0){
                            temp[initialState] = 0.0;
                        }
                        else{
                            temp[initialState] = 
                            beliefNodes[parentIndex][site](rate,initialState)/
                                               contrib;
                        }
#ifdef DEBUG
                        assert( !isnan( temp[initialState] ) ) ;
                        assert( !isinf( temp[initialState] ) ) ;
#endif
                    }//end initialState
                
                    for (int finalState=0 ; finalState<numberStates; ++finalState){
                        double contrib = 0.0;
                        for (int initialState=0 ; initialState<numberStates;
                             ++initialState){
                            contrib += lookup(rate, initialState, finalState ) *
                                       temp[initialState] ;
                        }
    	                beliefNodes[i][site](rate, finalState) = contrib *
          nodeRefVector[i]->partialLikelihood[modelId](site, finalState, rate) ;
#ifdef DEBUG
                        assert( !isnan( beliefNodes[i][site](rate, finalState) ) ) ;
                        assert( !isinf( beliefNodes[i][site](rate, finalState) ) ) ;
#endif
                    } // finalState
                } //sites
            } // nodes
        } // rates
        
        // calculate priors
        for ( int rate = 0; rate < numberRatesCategories; ++rate ) {
            for ( int i = 0; i < getNumberNodes(); ++i ){
                int numberChildren = ( nodeRefVector[i]->child[2] == NULL ) ?
                                     2 : 3;

                for ( int branch = 0; branch < numberChildren; ++branch ) {
                    int r = (nodeRefVector[i]->child[branch]->isLeaf()) ?
                            0 : rate;
                    array3D < double > lookup( numberRatesCategories,
                                               numberStates, numberStates );
                    double branchLength = nodeRefVector[i]->cdistance[branch];
                    // Initialize lookup table
                    for ( int initialState = 0; initialState<numberStates;
                          ++initialState){
                        for ( int finalState = 0; finalState < numberStates;
                              ++finalState ){
                            lookup( rate, initialState, finalState ) =
                                t_model->probability(  initialState, finalState,
                                                       branchLength, rate,
                                                       modelId );
                        }
                    }
                    for ( unsigned int site = 0; site < totalLength; ++site ) {
                        for ( int initialState = 0; initialState < numberStates;
                              ++initialState ) {
                            double contrib = 0.0;
                            for ( int finalState = 0; finalState < numberStates;
                                  ++finalState ){
                                contrib +=
                                     lookup(rate, initialState, finalState ) *
              nodeRefVector[i]->child[branch]->partialLikelihood[modelId]
                                                         (site, finalState, r);
                                
                            } // finalState

                            if ( beliefNodes[i][site](rate, initialState) ==
                                 0.0 ){
                                if (nodeRefVector[i]->child[branch]->isLeaf()){
                                    priorLeaves[nodeRefVector[i]->
                                            child[branch]->index]
                                             [site]( rate, initialState ) = 0.0;
                                }
                                else{
                                    priorNodes[nodeRefVector[i]->
                                            child[branch]->index]
                                          [site]( rate, initialState ) = 0.0;
                                }
                            }
                            else{
                                if (nodeRefVector[i]->child[branch]->isLeaf()){
                                    priorLeaves[nodeRefVector[i]->
                                        child[branch]->index]
                                      [site]( rate, initialState ) =
                                    beliefNodes[i][site](rate, initialState)/
                                          contrib;
#ifdef DEBUG
                                    assert( !isnan( priorLeaves[
                                        nodeRefVector[i]->child[branch]->index]
                                          [site]( rate, initialState ) ) ) ;
                                    assert( !isinf( priorLeaves[
                                        nodeRefVector[i]->child[branch]->index]
                                          [site]( rate, initialState ) ) ) ;
#endif
                                }
                                else{
                                    priorNodes[nodeRefVector[i]->
                                            child[branch]->index]
                                              [site]( rate, initialState ) =
                                    beliefNodes[i][site](rate, initialState)/
                                                  contrib;
#ifdef DEBUG
                                    assert( !isnan( priorNodes[
                                        nodeRefVector[i]->child[branch]->index]
                                          [site]( rate, initialState ) ) ) ;
                                    assert( !isinf( priorNodes[
                                        nodeRefVector[i]->child[branch]->index]
                                          [site]( rate, initialState ) ) ) ;
#endif
                                }
                            }
                        } // initialState
                    } //site
                } // branch
            } // nodes
        } //rate
        
        // calculate the gradients
        for ( int i = 0; i < getNumberBranches(); ++i ) {
            int nodeId;
            int branchId;
            retrieveBranchEnds( i, nodeId, branchId );
            double contrib = nodeRefVector[nodeId]->cdistance[branchId];
            // Initialise lookup tables
            array3D < double > difflookup( numberRatesCategories, numberStates,
                                           numberStates );
            array3D < double > lookup( numberRatesCategories, numberStates,
                                       numberStates );

            for (int rate = 0; rate < numberRatesCategories; ++rate){
                for ( int initialState = 0; initialState < numberStates;
                      ++initialState ){
                    for ( int finalState = 0; finalState < numberStates;
                          ++finalState ){
                        lookup( rate, initialState, finalState ) =
                            t_model->probability( initialState, finalState,
                                                  contrib, rate, modelId );
                        difflookup( rate, initialState, finalState ) =
                            t_model->diffProbability( initialState, finalState,
                                                      contrib, rate, modelId );
                    } //end finalState
                } //end initialState
            } // end rate

            for ( unsigned int site = 0; site < totalLength; ++site){
                double siteLikelihood = 0.0;
                double siteGradient = 0.0;
                for ( int rate = 0; rate < numberRatesCategories; ++rate ){
                    double sumdiff = 0.0;
                    double suml = 0.0;             
                    if ( ! nodeRefVector[nodeId]->child[branchId]->isLeaf() ){
                        for ( int initialState = 0; initialState < numberStates;
                          ++initialState ){
                            for ( int finalState = 0; finalState < numberStates;
                                  ++finalState ){
                                int index = nodeRefVector[nodeId]->
                                                          child[branchId]->index;
                                sumdiff += difflookup(rate, initialState,
                                                                finalState) *
                                     priorNodes[index][site](rate, initialState)*
                                     nodeRefVector[nodeId]->child[branchId]->
                                          partialLikelihood[modelId]( site, finalState, rate );
                                suml += lookup(rate, initialState, finalState) *
                                     priorNodes[index][site](rate, initialState)*
                                     nodeRefVector[nodeId]->child[branchId]->
                                          partialLikelihood[modelId]( site, finalState, rate );
                            } // end final state
                        } // end initial state
                    }
                    //if the child is a leaf use prior for the leaves 
                    //and use the unique partial likelihood
                    else{
                        for ( int initialState = 0; initialState < numberStates;
                          ++initialState ){
                            for ( int finalState = 0; finalState < numberStates;
                                  ++finalState ){
                                int index = nodeRefVector[nodeId]->
                                                        child[branchId]->index;
                                sumdiff += difflookup(rate, initialState,
                                                               finalState) *
                                     priorLeaves[index][site](rate, initialState)*
                                     nodeRefVector[nodeId]->child[branchId]->
                                          partialLikelihood[modelId]( site, finalState, 0 );
                                suml += lookup(rate, initialState, finalState) *
                                     priorLeaves[index][site](rate, initialState)*
                                     nodeRefVector[nodeId]->child[branchId]->
                                          partialLikelihood[modelId]( site, finalState, 0 );
                            } // end final state
                        } // end initial state
                    }
                    siteGradient += t_model->
                            getRateCategoryProbability(rate, modelId) * sumdiff;
                    siteLikelihood += t_model->
                            getRateCategoryProbability(rate, modelId) * suml;
#ifdef DEBUG
                    assert( !isnan( siteGradient ) );
                    assert( !isinf( siteGradient ) );
                    assert( !isnan( siteLikelihood ) );
                    assert( !isinf( siteLikelihood ) );
#endif
                } // end rate
                if (site<sequencesLength){
                    gradientVector[i] += ( 1.0 / siteLikelihood * siteGradient );
                }
                else{
                    gradientVector[i] += (1.0 / siteLikelihood * siteGradient) *
             t_table->getInvariantBases(modelId)[site-sequencesLength].second;
                }
            } // site
        } //end branches
    } // end modelId
#ifdef DEBUG
    for (unsigned int i = 0; i < gradientVector.size(); ++i){
        assert( !isnan( gradientVector[i] ) );
        assert( !isinf( gradientVector[i] ) );

    }
#endif
    return (gradientVector);    
}

int InferenceTree::normalize( int branch, int left, int right ) {

    int branchFromNode, branchSide;

    int swap1, swap2;

    assert( ( left == 0 ) || ( left == 1 ) );
    assert( ( right == 0 ) || ( right == 1 ) );
    if ( branch >= ( getNumberBranches() - 3 ) ) {
        branchFromNode = getNumberNodes() - 1;
        branchSide = branch - getNumberBranches() + 3;
    }
    else {
        branchFromNode = branch / 2;
        branchSide = ( branch % 2 );
    }

    assert( nodeRefVector[branchFromNode]->cdistance[branchSide] <= 0.0 );

    //if the negative branch is leading to a leaf then "rebound" :
    //branchLength=abs(branchLength)
    if ( nodeRefVector[branchFromNode]->child[branchSide]->isLeaf() ) {
        nodeRefVector[branchFromNode]->cdistance[branchSide] =
        fabs( nodeRefVector[branchFromNode]->cdistance[branchSide] );
        nodeRefVector[branchFromNode]->child[branchSide]->pdistance =
        nodeRefVector[branchFromNode]->cdistance[branchSide];
        return 0;
    }

    //if the negative branch is leading to an internal node
    //root the tree with the from node to allow computation.
    if ( nodeRefVector[branchFromNode] != root ) {
        nodeRefVector[branchFromNode]->makeRoot();
        root = nodeRefVector[branchFromNode];
        createIndex();
        branchFromNode = getNumberNodes() - 1;
    }

    swap1 = left;
    swap2 = ( branchSide + 1 + ( right ) ) % 3;


    nodeRefVector[branchFromNode]->cdistance[branchSide] =
    fabs( nodeRefVector[branchFromNode]->cdistance[branchSide] );
    nodeRefVector[branchFromNode]->child[branchSide]->pdistance =
    nodeRefVector[branchFromNode]->cdistance[branchSide];
    assert( root->NNI( branchSide, swap1, swap2 ) );
    createIndex();
    return 1;
}

/*
int tree::normalize() {
int i,j , swap ;
bool change = false ;


for (i=0 ; i<nnodes ; i++) {
for (j=0 ; j<3 ; j++) {
if (node[i]->child[j] != NULL) {
if (node[i]->cdistance[j] < 0.0) {	
if (node[i]->child[j]->isLeaf()) {
node[i]->calculated = false ;
node[i]->cdistance[j] = node[i]->child[j]->pdistance =
fabs(node[i]->cdistance[j]) ;
}
else {	
change = true ;		
if (node[i] != root)
swap = (j==0 ? 1 : 0) ;
else
swap = (j + 1 + (int)(floor(2.0 * ran1()))) % 3 ;	
node[i]->cdistance[j] = node[i]->child[j]->pdistance =
fabs(node[i]->cdistance[j]) ;
assert(j != swap) ;
node[i]->NNI(j , (int)(floor(2.0 * ran1())) , swap) ;	    	
}
}
}
}
}

if (change) {
create_index() ;
return 1 ;
}
return 0 ;
}
*/



bool InferenceTree::swap_distance_two( int branchId, bool up, int coinFlip, int rand1,
int rand2 ) {
    assert( 0 );
    branchId = branchId;
    up = up;
    coinFlip = coinFlip;
    rand1 = rand2;

    /*
    int fromNode;
    int toNode;
    int child2;

    int numberBranches = getNumberBranches();

    Node * temp;

    if ( branchId >= ( numberBranches - 3 ) ) {
    fromNode = getNumberNodes() - 1;
    toNode = branchId - numberBranches + 3;
    if ( !up ) {
    temp = root->child[toNode];
    temp->makeRoot();
    root = temp;
    createIndex();
    return ( swap_distance_two( nbranches - 1, true, coin_flip,
    rand1, rand2 ) );
    }
    }
    else {
    fromNode = branchId / 2;
    toNode = ( branchId % 2 );
    if ( up ) {
    nodeRefVector[fromNode]->makeRoot();
    root = nodeRefVector[fromNode];
    createIndex();
    fromNode = getNumberNodes() - 1;
    }
    else{
    temp = nodeRefVector[fromNode]->child[toNode];
    temp->makeRoot();
    root = temp;
    createIndex();
    return ( swap_distance_two( nbranches - 1, true, coin_flip,
    rand1, rand2 ) );
    }
    }
    ----------------------------
    if ( root->child[( child1 + 1 ) % 3]->isLeaf() &&
    root->child[( child1 + 2 ) % 3]->isLeaf() ) {
    // cerr << "Fault Here" << endl ;
    temp = root->child[child1];
    temp->makeRoot();
    root = temp;
    create_index();
    node1 = nnodes - 1;
    child1 = 2;
    if ( root->child[0]->isLeaf() && root->child[1]->isLeaf() )
    return false;
    if ( root->child[0]->isLeaf() && !root->child[1]->isLeaf() ) child2 = 1;
    if ( !root->child[0]->isLeaf() && root->child[1]->isLeaf() ) child2 = 0;
    if ( !root->child[0]->isLeaf() && !root->child[1]->isLeaf() )
    child2 = coin_flip % 2;


    swap_branches( ( root->child[child2]->index * 2 + ( rand2 % 2 ) ),
    ( root->child[child1]->index * 2 ) + ( rand1 % 2 ) );


    return true;
    }

    if ( direction ) {
    // cerr << "Fault Here 2 : " << child1 << endl ;
    if ( !root->child[( child1 + 1 ) % 3]->isLeaf() &&
    root->child[( child1 + 2 ) % 3]->isLeaf() ) {
    // cerr << "First" << endl ;
    cerr << toStringNumbered( false ) << endl;
    child2 = ( child1 + 1 ) % 3;
    // cerr << "Child2 = " << child2 << endl ;
    }
    if ( root->child[( child1 + 1 ) % 3]->isLeaf() &&
    !root->child[( child1 + 2 ) % 3]->isLeaf() ) {
    // cerr << "Second" << endl ;
    child2 = ( child1 + 2 ) % 3;
    }
    if ( !root->child[( child1 + 1 ) % 3]->isLeaf() &&
    !root->child[( child1 + 2 ) % 3]->isLeaf() ) {
    //  cerr << "Double" << endl ;
    child2 = ( child1 + ( ( coin_flip + 1 ) % 3 ) ) % 3;
    }
    swap_branches( ( root->child[child2]->index * 2 + ( rand2 % 2 ) ),
    ( root->child[child1]->index * 2 ) + ( rand1 % 2 ) );
    //  swap_branches((root->child[child2]->index*nbranches+(rand2%2))
    // , (root->child[child1]->index*nbranches)+(rand1%2));


    return true;
    }

    */

    return false;
}




