#include "Node.h"

#include <stdio.h>
#include <assert.h>

#include <iostream>

#include "Models/Model.h"

#include "Tree/depthmap.h"

#include "Util/array2D.h"

#include "statlib.h"

#include "PatternDesign/Singleton.h"
#include "Util/randombox.h"

#include "Tree/Node.h"

#include "Tree/Tree.h"

class SequenceTable;

Node * Node::makenode( string & nodeString, Tree * ptree  ) {

    Node * theNode;
    bool leaf = true;
    for ( unsigned int i = 0;  i < nodeString.length(); ++i ) {
        if ( nodeString[i] == '(' ) {
            leaf = false;
        }
    }
    if ( leaf ) {
        theNode = new Node( nodeString, ptree );
    }
    else {
        depthmap map( nodeString );
        theNode = new Node( ptree );
        int n = 0;
        for ( int i = 1; i < map.length; i++ ) {
            if ( map.depths[0] == ( map.depths[i] - 1 ) ) {
                theNode->child[n] = Node::create_node( map, i, ptree );
                theNode->child[n]->parent = theNode;
                theNode->child[n]->pdistance = theNode->cdistance[n] = map.distance[i];
                ++n;
            }
        }
        if ( ptree->t_table ) {
            assert( ptree->t_model );
            theNode->computeIdentical();
        }
    }
    return theNode;
}

// Function the recursively creates a tree structure using the inputted
// depthmap variable map.
Node * Node::create_node( depthmap & map, int start, Tree * ptree   ) {
    Node * rnode;

    int i = start + 1, n = 0;

    // If the depth map indicates a leaf
    if ( map.identifier[start] ) {
        rnode = new Node( map.label[start], ptree );
        return rnode;
    }

    //rnode = new Node( map.label[start], ptree );

    rnode = new Node( ptree );

    while ( ( n < 2 ) && ( i < map.length ) ) {
        if ( ( map.depths[start] + 1 ) == map.depths[i] ) {
            rnode->child[n] = create_node( map, i, ptree );
            rnode->child[n]->parent = rnode;
            rnode->child[n]->pdistance = rnode->cdistance[n] = map.distance[i];
            ++n;
        }
        ++i;
    }
    if ( ptree->t_table ) {
        assert( ptree->t_model );
        rnode->computeIdentical();
    }
    return ( rnode );
}

/** ****************************************************************************
 * Node Implementation
 **************************************************************************** */

// Constructors
Node::Node() {
    parent = NULL;
    for ( int i = 0; i < 3; ++i ) {
        child[i] = NULL;
    }
    calculated = false;
    clustered = false;
    index = 200000000;
    cnumber = -1;
    ptree = NULL;
}

Node::Node( Tree * ptree ) {
    this->ptree = ptree;
    label = string( "" );
    pdistance = 0.0;
    parent = NULL;
    clustered = false;
    calculated = false;
    for ( int i = 0; i < 3; i++ ) {
        cdistance[i] = 0.0;
        child[i] = NULL;
        lookup[i] = NULL;
    }
    if ( ptree->t_table ) {
        assert( ptree->t_model );
        initialisation( false );
    }
    index = 20000000;
}


Node::Node( string label, Tree * ptree  ) {
    this->ptree = ptree;
    this->label = label;
    pdistance = 0.0;
    parent = NULL;
    clustered = false;
    calculated = false;
    for ( int i = 0; i < 3; i++ ) {
        cdistance[i] = 0.0;
        child[i] = NULL;
        lookup[i] = NULL;
    }
    index = 20000000;
    if ( ptree->t_table ) {
        assert( ptree->t_model );
        initialisation( true );
    }
}

void Node::initialisation( bool leafNode ) {

    identical.resize( ptree->t_table->getNumberModels() );
    lastCalc.resize( ptree->t_table->getNumberModels() );
    partialLikelihood.resize( ptree->t_table->getNumberModels() );

    for( unsigned int modelId = 0; modelId < ptree->t_table->getNumberModels();
         ++modelId ){
        const array2D<string> & sequences =
                ptree->t_table->getSequences( modelId );
        const vector< pair< string, unsigned int > > & invariants =
            ptree->t_table->getInvariantBases( modelId );
        int size = sequences.numberColumns() + invariants.size();
        identical[modelId].resize(size);
        lastCalc[modelId].resize(ptree->t_model->getNumberSymbols(modelId));
    }

    if ( leafNode ) {
        int sequenceId = ptree->t_table->findsequence( label );
        if ( sequenceId == -1 ) {
            cerr << "Unable to find sequence for \"" << label << "\"" << endl;
            exit( 1 );
        }
        cnumber = sequenceId + 1;
        for( unsigned int modelId = 0;
             modelId < ptree->t_table->getNumberModels();
             ++modelId ){
            const array2D<string> & sequences =
                    ptree->t_table->getSequences( modelId );
            const vector< pair< string, unsigned int > > & invariants =
                ptree->t_table->getInvariantBases( modelId );
            int size = sequences.numberColumns() + invariants.size();
            partialLikelihood[modelId].resize(size,
                                     ptree->t_model->getNumberStates(modelId),
                                     1 );
            for ( unsigned int site = 0; site < sequences.numberColumns();
                  ++site ) {
                identical[modelId][site] =
                    ptree->t_model->getSymbolNumber( sequences(sequenceId,site), modelId );
                if ( identical[modelId][site] ==
                    ptree->t_model->getNumberSymbols( modelId ) ){
                    cerr << "unrecognized symbol " << label << " " << sequences(sequenceId,site) << endl;
                    vector < vector< unsigned int > > pos =
                        ptree->t_table->retrieveInitialNucleotides( modelId, site );
                    for ( vector < vector< unsigned int > >::iterator iter = pos.begin();
                          iter != pos.end(); ++iter ){
                        for ( vector< unsigned int >::iterator iter2 = (*iter).begin();
                              iter2 != (*iter).end(); ++iter2 ){
                             if (iter2 != (*iter).begin() ){
                                 cout << ", ";
                             }
                             cout << *iter2 + 1;
                        }
                        cout << "; ";
                    }
                    cout << endl;
                    exit(1);
                }
            }
            for ( unsigned int inv = 0; inv< invariants.size(); ++inv ) {
                identical[modelId][inv+sequences.numberColumns()] =
                    ptree->t_model->getSymbolNumber( invariants[inv].first, modelId );
                if ( identical[modelId][inv+sequences.numberColumns()] ==
                    ptree->t_model->getNumberSymbols( modelId ) ){
                    cerr << "unrecognized symbol " << invariants[inv].first << endl;
                    vector < vector< unsigned int > > pos =
                        ptree->t_table->retrieveInitialNucleotides( modelId, -inv-1);
                    for ( vector < vector< unsigned int > >::iterator iter = pos.begin();
                          iter != pos.end(); ++iter ){
                        for ( vector< unsigned int >::iterator iter2 = (*iter).begin();
                              iter2 != (*iter).end(); ++iter2 ){
                             if (iter2 != (*iter).begin() ){
                                 cout << ", ";
                             }
                             cout << *iter2 + 1;
                        }
                        cout << "; ";
                    }
                    cout << endl;
                    exit(1);
                }
            }
        }
        for ( int childId = 0; childId < 3; ++childId ) {
            lookup[childId]=NULL;
        }
    }
    //if !leafNode
    else {
        for( unsigned int modelId = 0;
             modelId < ptree->t_table->getNumberModels(); ++modelId ){
            int size = ptree->t_table->getSequencesLength(modelId) +
                       ptree->t_table->getInvariantsLength(modelId);
            partialLikelihood[modelId].resize(size,
                            ptree->t_model->getNumberStates(modelId),
                            ptree->t_model->getNumberRateCategories(modelId) );
        }
        //create lookup table
        for ( int childId = 0; childId < 3; ++childId ) {
            lookup[childId] = new array3D < double >
                    [ptree->t_table->getNumberModels()];
            for( unsigned int modelId = 0;
                 modelId < ptree->t_table->getNumberModels(); ++modelId ){
                lookup[childId] [modelId].resize(
                    ptree->t_model->getNumberRateCategories(modelId),
                    ptree->t_model->getNumberStates(modelId),
                    ptree->t_model->getNumberStates(modelId) );
            }
        }
    }
}


// Destructor
Node::~Node() {
    destroy();
}

string Node::toStringClade( vector < int > * claderep, int numclades ) {
    string tree;
    int i;

    //if clade in claderep
    for ( i = 0; i < numclades; i++ ) {
        if ( claderep[i] == clade ) {
            tree = string( "A" );
            tree[0] += i;
            return tree;
        }
    }

    //if leaf node
    if ( isLeaf() ) {
        return label;
    }

    //otherwise
    tree = "(";
    if ( child[0] != NULL ) {
        tree += child[0]->toStringClade( claderep, numclades );
    }
    if ( child[1] != NULL ) {
        tree += "," + child[1]->toStringClade( claderep, numclades );
    }
    if ( child[2] != NULL ) {
        tree += "," + child[2]->toStringClade( claderep, numclades );
    }
    tree += ")";
    return tree;
}


string Node::toString() {
    string tree;
    char d1[64], d2[64], d3[64];;

    //if leaf node
    if ( isLeaf() ) {
        return label;
    }

    sprintf( d1, "%0.4f", cdistance[0] );
    sprintf( d2, "%0.4f", cdistance[1] );
    sprintf( d3, "%0.4f", cdistance[2] );

    tree = "(";
    if ( child[0] != NULL ) {
        tree += child[0]->toString() + ":" + d1;
    }
    if ( child[1] != NULL ) {
        tree += "," + child[1]->toString() + ":" + d2;
    }
    if ( child[2] != NULL ) {
        tree += "," + child[2]->toString() + ":" + d3;
    }
    tree += ")";
    return tree;
}

string Node::toStringTopology() {
    string tree;
    // if leaf node
    if ( isLeaf() ) {
        return label;
    }


    tree = "(";
    if ( child[0] != NULL ) {
        tree += child[0]->toStringTopology();
    }
    if ( child[1] != NULL ) {
        tree += "," + child[1]->toStringTopology();
    }
    if ( child[2] != NULL ) {
        tree += "," + child[2]->toStringTopology();
    }
    tree += ")";
    return tree;
}

string Node::toStringNumbered() {
    string tree;

    //if leaf node
    if ( isLeaf() ) {
        char d1[10];
        sprintf( d1, "%i", cnumber );
        return string( d1 );
    }

    tree = "(";
    if ( child[0] != NULL ) {
        tree += child[0]->toStringNumbered();
    }
    if ( child[1] != NULL ) {
        tree += "," + child[1]->toStringNumbered();
    }
    if ( child[2] != NULL ) {
        tree += "," + child[2]->toStringNumbered();
    }
    tree += ")";
    return tree;
}


// Primitive Node methods
void Node::destroy() {

    for ( int i = 0; i < 3; ++i ) {
        if ( child[i] != NULL ) {
            delete child[i];
        }
    }

    //delete lookup table 
    for ( int childId = 0; childId < 3; ++childId ) {
        delete[] lookup[childId];
    }
}

// Duplicates only the values at the source node to the target node
void Node::duplicate( Node & src ) {
    assert( 0 );
    src = src;

    /*
    int i;
    label = src.label;

    ptree = src.ptree
    calculated = src.calculated;
    clustered = src.clustered;
    cnumber = src.cnumber;

    if ( src.identical != NULL ) {
    if ( identical != NULL ){
    delete identical;
    }
    identical = new int[loaded_data->getSequencesLength()];
    memcpy( identical, src.identical,
    sizeof( int ) * loaded_data->getSequencesLength() );
    }

    // Copy internal arrays
    plikelihood = src.plikelihood;
    prior = src.prior;
    lookup = src.lookup;
    pdistance = src.pdistance;
    for ( i = 0; i < 3; i++ ) cdistance[i] = src.cdistance[i];
    */
}


// Copies nearly everything from source node to current node
// the reference to the tree (ptree, index,... are omitted)
// ptree of the parent is used, in the case parent==NULL
// the actual ptree is used for all the children
void Node::recursiveCopy( Node * srcNode, Node * parent ) {

    for ( int i = 0; i < 3; ++i ) {
        assert( child[i] == NULL );
    }

    label = srcNode->label;
    this->parent = parent;
    pdistance = srcNode->pdistance;
    if ( parent != NULL ) {
        ptree = parent->ptree;
    }
    assert( srcNode->ptree->t_table == ptree->t_table );
    assert( srcNode->ptree->t_model == ptree->t_model );

    calculated = srcNode->calculated;
    clustered = srcNode->clustered;
    cnumber = srcNode->cnumber;

    // Copy internal arrays
    if ( ( ptree->t_table != NULL ) || ( ptree->t_model != NULL ) ) {
        identical = srcNode->identical;
        if (calculated){
            partialLikelihood = srcNode->partialLikelihood;
        }
        else{
            partialLikelihood.resize( ptree->t_table->getNumberModels() );
            for( unsigned int modelId = 0;
                 modelId < ptree->t_table->getNumberModels(); ++modelId ){
                const array2D<string> & sequences =
                        ptree->t_table->getSequences( modelId );
                const vector< pair< string, unsigned int > > & invariants =
                    ptree->t_table->getInvariantBases( modelId );
                int size = sequences.numberColumns() + invariants.size();
                if (srcNode->isLeaf()){
                    partialLikelihood[modelId].resize(size,
                                       ptree->t_model->getNumberStates(modelId),
                                       1 );
                }
                else{
                    partialLikelihood[modelId].resize(size,
                            ptree->t_model->getNumberStates(modelId),
                            ptree->t_model->getNumberRateCategories(modelId) );
                }
            }
        }

        //create lookup table
        for ( int childId = 0; childId < 3; ++childId ) {
            lookup[childId] = new array3D < double >[ptree->t_table->getNumberModels()];
            for( unsigned int modelId = 0;
                 modelId < ptree->t_table->getNumberModels(); ++modelId ){
                lookup[childId] [modelId].resize(
                    ptree->t_model->getNumberRateCategories(modelId),
                    ptree->t_model->getNumberStates(modelId),
                    ptree->t_model->getNumberStates(modelId) );
            }
        }

    }     //end if ( ptree->t_table || ptree->t_model )

    lastCalc.resize( ptree->t_table->getNumberModels() );
    for( unsigned int modelId = 0; modelId < ptree->t_table->getNumberModels();
         ++modelId ){
        lastCalc[modelId].resize(ptree->t_model->getNumberSymbols(modelId));
    }

    for ( int i = 0; i < 3; i++ ) {
        if ( ( srcNode->child[i] != NULL ) ) {
            child[i] = new Node();
            cdistance[i] = srcNode->cdistance[i];
            child[i]->recursiveCopy( srcNode->child[i], this );
        }
    }
}


Node & Node::operator = ( Node & src ) {
    assert( 0 );
    src = src;

    /*
    for ( int i = 0; i < 3; i++ ) {
    if ( ( src.child[i] == child[i] ) && ( child[i] != NULL ) ){
    return ( src );
    }
    copy( src );
    */

    return ( * this );
}

// Nearest neighbour interchange
bool Node::NNI( int branch, int abranch, int cbranch ) {

    Node * temp;
    double tempd;

    if ( isLeaf() || ( child[branch]->isLeaf() ) ) {
        assert( 0 );
    }

    if ( branch == cbranch ) {
        assert( 0 );
    }

    temp = child[branch]->child[abranch];
    tempd = child[branch]->cdistance[abranch];

    child[branch]->child[abranch] = child[cbranch];
    child[branch]->child[abranch]->parent = child[branch];
    child[branch]->cdistance[abranch] = cdistance[cbranch];
    child[branch]->child[abranch]->pdistance = cdistance[cbranch];

    child[cbranch] = temp;
    temp->parent = this;
    cdistance[cbranch] = temp->pdistance = tempd;

    calculated = false;
    child[branch]->calculated = false;

    temp = parent;
    while ( temp != NULL ) {
        temp->calculated = false;
        temp = temp->parent;
    }

    if ( ptree->t_table != NULL ) {
        child[branch]->computeIdentical();
    }
    return true;
}


void Node::computeIdentical() {
    unsigned int modelId;

    assert(ptree->t_table != NULL);
    
    if ( !isLeaf() ) {
#ifdef DEBUG
            for ( modelId=0; modelId<ptree->t_table->getNumberModels();
                  ++modelId){
                unsigned int length =
                        ptree->t_table->getSequencesLength(modelId);
                for ( unsigned int i = 0; i < length; ++i ) {
                    identical[modelId][i] = -1;
                }
                for ( unsigned int i = 0;
                      i < ptree->t_table->getInvariantsLength(modelId); ++i ) {
                    identical[modelId][length+i] = child[0]->identical[modelId][length+i];
                }
            }
        
#endif
        if ( child[2] == NULL ) {
            for ( modelId=0; modelId<ptree->t_table->getNumberModels();
                  ++modelId){
                unsigned int length =
                        ptree->t_table->getSequencesLength(modelId);
                for ( unsigned int i = 0; i < length; ++i ) {
                    if ( child[0]->identical[modelId][i] ==
                            child[1]->identical[modelId][i] ) {
                        identical[modelId][i] = child[0]->identical[modelId][i];
                    }
                    else{
                        identical[modelId][i] = -1;
                    }
                }
                for ( unsigned int i = 0;
                      i < ptree->t_table->getInvariantsLength(modelId); ++i ) {
                    identical[modelId][length+i] =
                            child[0]->identical[modelId][length+i];
                }
            }
        }
        else {
            //must compute identical for the root too (in case of a topology
            //movement) ?
            for ( modelId=0; modelId<ptree->t_table->getNumberModels();
                  ++modelId){
                unsigned int length =
                        ptree->t_table->getSequencesLength(modelId);
                for ( unsigned int i = 0; i < length; ++i ) {
                    identical[modelId][i] = -1;
                }
                for ( unsigned int i = 0;
                      i < ptree->t_table->getInvariantsLength(modelId); ++i ) {
                    identical[modelId][length+i] =
                            child[0]->identical[modelId][length+i];
                }
            }
        }
    }
    else {
        // identical for a leaf is invariant, error if trying to calculate
        // should have been initialised when the sequence is loaded
        // (loadLeafData method)
        assert( 0 );
    }
}


bool Node::simulate( const vector< vector<int> > & categories,
                     const vector< vector<int> > & parent,
                     vector< vector < vector<int> > > & result ) {

    Singleton < randombox > & randBox = Singleton < randombox >::instance();

    unsigned int numberModels = categories.size();
            
    vector< vector<int> > seq(numberModels);
    
    //if root node
    if (parent.empty()){
        assert(child[2] != NULL);
        for ( unsigned int modelId = 0; modelId < numberModels; ++modelId ) {
            for ( unsigned int site = 0; site < categories[modelId].size();
                  ++site ) {
               /* assign a frequency according to the probabilities for the
                  rate */
                int state = 0;
                int rate = categories[modelId][site];
                double number = randBox.ran() -
                       ptree->t_model->getFrequency( state, rate, modelId );
                while ( number > 0 ) {
                    ++state;
                    number -=
                        ptree->t_model->getFrequency( state, rate, modelId );
                }
                assert ( state < ptree->t_model->getNumberStates(modelId) );
                seq[modelId].push_back(state);
            }
        }
        for ( int childId = 0; childId < 3; ++childId ){
            child[childId]->simulate( categories, seq, result );
        }
    }
    else{
        assert(child[2] == NULL);
        for ( unsigned int modelId = 0; modelId < numberModels; ++modelId ) {
            for ( unsigned int site = 0; site < categories[modelId].size();
                  ++site ) {
                int state = 0;
                int rate = categories[modelId][site];
                double number = randBox.ran() -
                    ptree->t_model->probability( parent[modelId][site] ,state,
                                                 pdistance, rate, modelId );
                while ( number > 0 ){
                    ++state;
                    number -=  ptree->t_model->probability(
                       parent[modelId][site] ,state, pdistance, rate, modelId );
                }
                assert ( state < ptree->t_model->getNumberStates(modelId) );
                seq[modelId].push_back(state);
            }
        }
        if (isLeaf()){
            result.push_back( seq );
        }
        else{
            for ( int childId = 0; childId < 2; ++childId ){
                child[childId]->simulate( categories, seq, result );
            }
        }
    }
    return true;
}

void Node::codiff_likelihood( int diffbranch1, int diffbranch2 ) {
    assert( 0 );
    diffbranch1 = diffbranch2;

    /*
    // Make sure model and data have been loaded
    if ( ( loaded_data == NULL ) || ( loaded_model == NULL ) ) {
    if ( loaded_data == NULL ) {
    cerr << "The data has not been loaded" << endl;
    exit( 1 );
    }
    if ( loaded_model == NULL ) {
    cerr << "The model has not been loaded" << endl;
    exit( 1 );
    }
    }

    if ( isLeaf() ) return;

    int nucleotide, site, rate_cat, dnucleotide, j, i, branch;
    int nstates  = loaded_model->getNumberStates();
    int nrates   = loaded_model->getNumberRateCategories();
    int nsymbols = loaded_model->getNumberSymbols();
    int slength  = loaded_data->getSequencesLength();
    int last_calc[nsymbols];
    int r[3], nbranch = ( child[2] != NULL ? 3 : 2 );
    double lookup[nbranch] [nstates] [nstates], temp1, temp2, temp3;
    double * look1, * look2, * look3;
    plikelihood.resize( nrates, slength, nstates );



    for ( rate_cat = 0; rate_cat < nrates; rate_cat++ ) {
    for ( branch = 0; branch < 3; branch++ ) {
    if ( child[branch] != NULL )
    r[branch] = ( child[branch]->isLeaf() ? 0 : rate_cat );
    }
    for ( i = 0; i < nsymbols; i++ ) last_calc[i] = -1;

    // Intialise lookup table
    for ( branch = 0; branch < nbranch; branch++ )
    for ( i = 0; i < nstates; i++ )
    for ( j = 0; j < nstates; j++ ) {
    if ( ( branch != diffbranch1 ) && ( branch != diffbranch2 ) )
    lookup[branch] [i] [j] =
    loaded_model->probability( i, j,
    cdistance[branch], rate_cat );
    else
    lookup[branch] [i] [j] =
    loaded_model->diffProbability( i, j,
    cdistance[branch], rate_cat );
    }

    if ( nbranch == 2 ) {
    for ( site = 0; site < slength; site++ ) {
    if ( identical[site] < 0 ) {
    for ( nucleotide = 0; nucleotide <
    nstates; nucleotide++ ) {
    temp1 = temp2 = 0.0;
    look1 = lookup[0] [nucleotide];
    look2 = lookup[1] [nucleotide];
    for ( dnucleotide =
    0; dnucleotide < nstates; dnucleotide++ ) {
    temp1 += ( look1[dnucleotide] *
    child[0]->plikelihood( r[0], site,
    dnucleotide ) );
    temp2 += ( look2[dnucleotide] *
    child[1]->plikelihood( r[1], site,
    dnucleotide ) );
    }
    plikelihood( rate_cat, site, nucleotide ) = temp1 * temp2;
    }
    }
    else {
    if ( last_calc[identical[site]] >= 0 ) {
    j = last_calc[identical[site]];
    for ( nucleotide = 0; nucleotide < nstates; nucleotide++ )
    plikelihood( rate_cat, site, nucleotide ) =
    plikelihood( rate_cat, j, nucleotide );
    }
    else {
    for ( nucleotide = 0; nucleotide <
    nstates; nucleotide++ ) {
    temp1 = temp2 = 0.0;
    look1 = lookup[0] [nucleotide];
    look2 = lookup[1] [nucleotide];
    for ( dnucleotide =
    0; dnucleotide <
    nstates; dnucleotide++ ) {
    temp1 += ( look1[dnucleotide] *
    child[0]->plikelihood( r[0],
    site, dnucleotide ) );
    temp2 += ( look2[dnucleotide] *
    child[1]->plikelihood( r[1],
    site, dnucleotide ) );
    }
    plikelihood( rate_cat, site, nucleotide ) =
    temp1 * temp2;
    }
    last_calc[identical[site]] = site;
    } // else last_calc[...
    } // else identical[..
    } // for site
    } // if (nbranch ==
    else {
    for ( site = 0; site < slength; site++ ) {
    for ( nucleotide = 0; nucleotide < nstates; nucleotide++ ) {
    temp1 = temp2 = temp3 = 0.0;
    look1 = lookup[0] [nucleotide];
    look2 = lookup[1] [nucleotide];
    look3 = lookup[2] [nucleotide];
    for ( dnucleotide = 0; dnucleotide <
    nstates; dnucleotide++ ) {
    temp1 += ( look1[dnucleotide] *
    child[0]->plikelihood( r[0], site, dnucleotide ) );
    temp2 += ( look2[dnucleotide] *
    child[1]->plikelihood( r[1], site, dnucleotide ) );
    temp3 += ( look3[dnucleotide] *
    child[2]->plikelihood( r[2], site, dnucleotide ) );
    }
    plikelihood( rate_cat, site, nucleotide ) = temp1 * temp2 * temp3;
    }
    } // for site
    } // else nbranch==2
    } // for (rate_cat; ...

    calculated = true;
    return;
    */
}

void Node::difflikelihood( int diffbranch ) {
    assert( 0 );
    diffbranch = diffbranch;

    /*
    // Make sure model and data have been loaded
    if ( ( loaded_data == NULL ) || ( loaded_model == NULL ) ) {
    if ( loaded_data == NULL ) {
    cerr << "The data has not been loaded" << endl;
    exit( 1 );
    }
    if ( loaded_model == NULL ) {
    cerr << "The model has not been loaded" << endl;
    exit( 1 );
    }
    }

    if ( isLeaf() ) return;

    int nucleotide, site, rate_cat, dnucleotide, j, i, branch;
    int nstates  = loaded_model->getNumberStates();
    int nrates   = loaded_model->getNumberRateCategories();
    int nsymbols = loaded_model->getNumberSymbols();
    int slength  = loaded_data->getSequencesLength();
    int last_calc[nsymbols];
    int r[3], nbranch = ( child[2] != NULL ? 3 : 2 );
    double lookup[nbranch] [nstates] [nstates], temp1, temp2, temp3;
    double * look1, * look2, * look3;
    plikelihood.resize( nrates, slength, nstates );



    for ( rate_cat = 0; rate_cat < nrates; rate_cat++ ) {
    for ( branch = 0; branch < 3; branch++ ) {
    if ( child[branch] != NULL )
    r[branch] = ( child[branch]->isLeaf() ? 0 : rate_cat );
    }
    for ( i = 0; i < nsymbols; i++ ) last_calc[i] = -1;

    // Intialise lookup table
    for ( branch = 0; branch < nbranch; branch++ )
    for ( i = 0; i < nstates; i++ )
    for ( j = 0; j < nstates; j++ ) {
    if ( branch != diffbranch )
    lookup[branch] [i] [j] =
    loaded_model->probability( i, j,
    cdistance[branch], rate_cat );
    else
    lookup[branch] [i] [j] =
    loaded_model->diffProbability( i, j,
    cdistance[branch], rate_cat );
    }

    if ( nbranch == 2 ) {
    for ( site = 0; site < slength; site++ ) {
    if ( identical[site] < 0 ) {
    for ( nucleotide = 0; nucleotide <
    nstates; nucleotide++ ) {
    temp1 = temp2 = 0.0;
    look1 = lookup[0] [nucleotide];
    look2 = lookup[1] [nucleotide];
    for ( dnucleotide =
    0; dnucleotide < nstates; dnucleotide++ ) {
    temp1 += ( look1[dnucleotide] *
    child[0]->plikelihood( r[0], site,
    dnucleotide ) );
    temp2 += ( look2[dnucleotide] *
    child[1]->plikelihood( r[1], site,
    dnucleotide ) );
    }
    plikelihood( rate_cat, site, nucleotide ) = temp1 * temp2;
    }
    }
    else {
    if ( last_calc[identical[site]] >= 0 ) {
    j = last_calc[identical[site]];
    for ( nucleotide = 0; nucleotide < nstates; nucleotide++ )
    plikelihood( rate_cat, site, nucleotide ) =
    plikelihood( rate_cat, j, nucleotide );
    }
    else {
    for ( nucleotide = 0; nucleotide <
    nstates; nucleotide++ ) {
    temp1 = temp2 = 0.0;
    look1 = lookup[0] [nucleotide];
    look2 = lookup[1] [nucleotide];
    for ( dnucleotide =
    0; dnucleotide <
    nstates; dnucleotide++ ) {
    temp1 += ( look1[dnucleotide] *
    child[0]->plikelihood( r[0],
    site, dnucleotide ) );
    temp2 += ( look2[dnucleotide] *
    child[1]->plikelihood( r[1],
    site, dnucleotide ) );
    }
    plikelihood( rate_cat, site, nucleotide ) =
    temp1 * temp2;
    }
    last_calc[identical[site]] = site;
    } // else last_calc[...
    } // else identical[..
    } // for site
    } // if (nbranch ==
    else {
    for ( site = 0; site < slength; site++ ) {
    for ( nucleotide = 0; nucleotide < nstates; nucleotide++ ) {
    temp1 = temp2 = temp3 = 0.0;
    look1 = lookup[0] [nucleotide];
    look2 = lookup[1] [nucleotide];
    look3 = lookup[2] [nucleotide];
    for ( dnucleotide = 0; dnucleotide <
    nstates; dnucleotide++ ) {
    temp1 += ( look1[dnucleotide] *
    child[0]->plikelihood( r[0], site, dnucleotide ) );
    temp2 += ( look2[dnucleotide] *
    child[1]->plikelihood( r[1], site, dnucleotide ) );
    temp3 += ( look3[dnucleotide] *
    child[2]->plikelihood( r[2], site, dnucleotide ) );
    }
    plikelihood( rate_cat, site, nucleotide ) = temp1 * temp2 * temp3;
    }
    } // for site
    } // else nbranch==2
    } // for (rate_cat; ...

    calculated = true;
    return;
    */
}

void Node::diff2likelihood( int diffbranch ) {
    assert( 0 );
    diffbranch = diffbranch;

    /*
    // Make sure model and data have been loaded
    if ( ( loaded_data == NULL ) || ( loaded_model == NULL ) ) {
    if ( loaded_data == NULL ) {
    cerr << "The data has not been loaded" << endl;
    exit( 1 );
    }
    if ( loaded_model == NULL ) {
    cerr << "The model has not been loaded" << endl;
    exit( 1 );
    }
    }

    if ( isLeaf() ) return;

    int nucleotide, site, rate_cat, dnucleotide, j, i, branch;
    int nstates  = loaded_model->getNumberStates();
    int nrates   = loaded_model->getNumberRateCategories();
    int nsymbols = loaded_model->getNumberSymbols();
    int slength  = loaded_data->getSequencesLength();
    int last_calc[nsymbols];
    int r[3], nbranch = ( child[2] != NULL ? 3 : 2 );
    double lookup[nbranch] [nstates] [nstates], temp1, temp2, temp3;
    double * look1, * look2, * look3;
    plikelihood.resize( nrates, slength, nstates );



    for ( rate_cat = 0; rate_cat < nrates; rate_cat++ ) {
    for ( branch = 0; branch < 3; branch++ ) {
    if ( child[branch] != NULL )
    r[branch] = ( child[branch]->isLeaf() ? 0 : rate_cat );
    }
    for ( i = 0; i < nsymbols; i++ ) last_calc[i] = -1;

    // Intialise lookup table
    for ( branch = 0; branch < nbranch; branch++ )
    for ( i = 0; i < nstates; i++ )
    for ( j = 0; j < nstates; j++ ) {
    if ( branch != diffbranch )
    lookup[branch] [i] [j] =
    loaded_model->probability( i, j,
    cdistance[branch], rate_cat );
    else
    lookup[branch] [i] [j] =
    loaded_model->secondDiffProbability( i, j,
    cdistance[branch], rate_cat );
    }

    if ( nbranch == 2 ) {
    for ( site = 0; site < slength; site++ ) {
    if ( identical[site] < 0 ) {
    for ( nucleotide = 0; nucleotide <
    nstates; nucleotide++ ) {
    temp1 = temp2 = 0.0;
    look1 = lookup[0] [nucleotide];
    look2 = lookup[1] [nucleotide];
    for ( dnucleotide =
    0; dnucleotide < nstates; dnucleotide++ ) {
    temp1 += ( look1[dnucleotide] *
    child[0]->plikelihood( r[0], site,
    dnucleotide ) );
    temp2 += ( look2[dnucleotide] *
    child[1]->plikelihood( r[1], site,
    dnucleotide ) );
    }
    plikelihood( rate_cat, site, nucleotide ) = temp1 * temp2;
    }
    }
    else {
    if ( last_calc[identical[site]] >= 0 ) {
    j = last_calc[identical[site]];
    for ( nucleotide = 0; nucleotide < nstates; nucleotide++ )
    plikelihood( rate_cat, site, nucleotide ) =
    plikelihood( rate_cat, j, nucleotide );
    }
    else {
    for ( nucleotide = 0; nucleotide <
    nstates; nucleotide++ ) {
    temp1 = temp2 = 0.0;
    look1 = lookup[0] [nucleotide];
    look2 = lookup[1] [nucleotide];
    for ( dnucleotide =
    0; dnucleotide <
    nstates; dnucleotide++ ) {
    temp1 += ( look1[dnucleotide] *
    child[0]->plikelihood( r[0],
    site, dnucleotide ) );
    temp2 += ( look2[dnucleotide] *
    child[1]->plikelihood( r[1],
    site, dnucleotide ) );
    }
    plikelihood( rate_cat, site, nucleotide ) =
    temp1 * temp2;
    }
    last_calc[identical[site]] = site;
    } // else last_calc[...
    } // else identical[..
    } // for site
    } // if (nbranch ==
    else {
    for ( site = 0; site < slength; site++ ) {
    for ( nucleotide = 0; nucleotide < nstates; nucleotide++ ) {
    temp1 = temp2 = temp3 = 0.0;
    look1 = lookup[0] [nucleotide];
    look2 = lookup[1] [nucleotide];
    look3 = lookup[2] [nucleotide];
    for ( dnucleotide = 0; dnucleotide <
    nstates; dnucleotide++ ) {
    temp1 += ( look1[dnucleotide] *
    child[0]->plikelihood( r[0], site, dnucleotide ) );
    temp2 += ( look2[dnucleotide] *
    child[1]->plikelihood( r[1], site, dnucleotide ) );
    temp3 += ( look3[dnucleotide] *
    child[2]->plikelihood( r[2], site, dnucleotide ) );
    }
    plikelihood( rate_cat, site, nucleotide ) = temp1 * temp2 * temp3;
    }
    } // for site
    } // else nbranch==2
    } // for (rate_cat; ...

    calculated = true;
    return;
    */
}

void Node::likelihood() {
    if ( calculated ) return;
  
    if ( isLeaf() ) {
        loadleafdata();
        calculated = true;
        return;
    }

    if ( ( child[2] == NULL ) && ( child[0]->isLeaf() ) &&
    ( child[1]->isLeaf() ) ) {
        optimizedLikelihoodLeafParent();
        calculated = true;
        return;
    }

    int numberChildren = ( child[2] != NULL ? 3 : 2 );

    double temp1, temp2, temp3;
    
    array3D<double> * look1;
    array3D<double> * look2;
    array3D<double> * look3;

    // if a children is a leaf there is no rate in the partial Likelyhood array
    // dim partialLikelyhood[site] = (numberStates for the site, 1)
    // and endRate[child] = 0.
    // otherwise dim partialLikelyhood[site] =
    //                (numberStates for the site, number rate for the site )
    // and "endRate = startRate"
    int endRate[numberChildren];

    // Initialize lookup table
    for ( unsigned int modelId = 0; modelId < ptree->t_table->getNumberModels();
          ++modelId ) {
        int numberStates = ptree->t_model->getNumberStates(modelId);
        int numberRateCategories =
                ptree->t_model->getNumberRateCategories(modelId);
        for ( int initialState = 0; initialState < numberStates; ++initialState ) {
            for ( int finalState = 0; finalState < numberStates; ++finalState ) {
               for ( int rate = 0; rate < numberRateCategories; ++rate ) {
                    for ( int child = 0; child < numberChildren; ++child ) {
                        lookup[child] [modelId](rate, initialState, finalState ) =
                        ptree->t_model->probability( initialState, finalState,
                                              cdistance[child], rate, modelId );
                    }
                }
            }
        }
        //lastCalc
        fill( lastCalc[modelId].begin(), lastCalc[modelId].end(), -1);
    }

    if ( numberChildren == 2 ) {
        for ( unsigned int modelId = 0;
              modelId < ptree->t_table->getNumberModels(); ++modelId ) {
            int sequenceLength = ptree->t_table->getSequencesLength(modelId);
            int invariantsLength = ptree->t_table->getInvariantsLength(modelId);
            int numberStates = ptree->t_model->getNumberStates(modelId);
            int numberRates =
                    ptree->t_model->getNumberRateCategories(modelId);
            look1 = &(lookup[0][modelId]);
            look2 = &(lookup[1][modelId]);
            for ( int site = 0; site < sequenceLength+invariantsLength; ++site ) {
                 if ( ( identical[modelId][site] == -1 ) ||
                     ( lastCalc[modelId][identical[modelId][site]] == -1 ) ){
                    for ( int rate = 0; rate < numberRates; ++rate ) {
                        endRate[0] = ( child[0]->isLeaf() ) ? 0 : rate;
                        endRate[1] = ( child[1]->isLeaf() ) ? 0 : rate;
                        for ( int initialState = 0; initialState < numberStates;
                              ++initialState ) {
                            temp1 = 0.0;
                            temp2 = 0.0;
                            for ( int finalState = 0; finalState < numberStates;
                                  ++finalState ) {
                                temp1 += ( (*look1)( rate, initialState,
                              finalState ) * child[0]->partialLikelihood[modelId]
                                    (site, finalState, endRate[0] ) );
                                temp2 += ( (*look2)( rate, initialState,
                              finalState ) * child[1]->partialLikelihood[modelId]
                                    (site, finalState, endRate[1] ) );
                            } // for each final state
                            partialLikelihood[modelId](site, initialState, rate) =
                                temp1 * temp2;
                        } // for each initial state
                    } // for each rate
                    // if identical[modelId][site] save this site ID for
                    // future computation
                    if ( identical[modelId][site] != -1 ) {
                        lastCalc[modelId] [identical[modelId][site]] = site;
                    }
                }
                // if identical[site]!=-1 && already calculated
                else {
                    memcpy( &(partialLikelihood[modelId](site,0,0)),
 &(partialLikelihood[modelId](lastCalc[modelId][identical[modelId][site]],0,0)),
                      sizeof( double ) * numberStates * numberRates );
#ifdef DEBUG
                    for ( int i = 0; i< numberStates; ++i ){
                        for ( int j = 0; j< numberRates; ++j ){
                            assert(partialLikelihood[modelId](site, i, j) ==
partialLikelihood[modelId](lastCalc[modelId][identical[modelId][site]], i, j) );
                        }
                    }
#endif
                   
                }
            } // for each site
        } // for each model
    } 
    // if it is the root node (numberChildren != 2)
    else {
#ifdef DEBUG
        assert(numberChildren==3);
#endif
        for ( unsigned int modelId = 0;
              modelId < ptree->t_table->getNumberModels(); ++modelId ) {
            int sequenceLength = ptree->t_table->getSequencesLength(modelId);
            int invariantsLength = ptree->t_table->getInvariantsLength(modelId);
            int numberStates = ptree->t_model->getNumberStates(modelId);
            int numberRates = ptree->t_model->getNumberRateCategories(modelId);
            look1 = & lookup[0][modelId];
            look2 = & lookup[1][modelId];
            look3 = & lookup[2][modelId];
            //identical is not useful anymore
            for ( int site = 0; site < sequenceLength+invariantsLength; ++site ) {
                for ( int rate = 0; rate < numberRates; ++rate ) {
                    endRate[0] = ( child[0]->isLeaf() ) ? 0 : rate;
                    endRate[1] = ( child[1]->isLeaf() ) ? 0 : rate;
                    endRate[2] = ( child[2]->isLeaf() ) ? 0 : rate;
                    for ( int initialState = 0; initialState < numberStates;
                          ++initialState ) {
                        temp1 = 0.0;
                        temp2 = 0.0;
                        temp3 = 0.0;
                        for ( int finalState = 0; finalState < numberStates;
                              ++finalState ) {
                            temp1 += ( ( * look1 ) ( rate, initialState,
                            finalState ) * child[0]->partialLikelihood[modelId]
                            (site, finalState, endRate[0] ) );
                            temp2 += ( ( * look2 ) ( rate, initialState,
                            finalState ) * child[1]->partialLikelihood[modelId]
                            (site, finalState, endRate[1] ) );
                            temp3 += ( ( * look3 ) ( rate, initialState,
                            finalState ) * child[2]->partialLikelihood[modelId]
                            (site, finalState, endRate[2] ) );
                        } // for each final state
                        partialLikelihood[modelId](site, initialState, rate) =
                            temp1 * temp2 * temp3;
                    } // for each initial state
                } // for each rate
            } // for each site
        } // for each model
    } // case of the root
    calculated = true;
}


void Node::loadleafdata() {

    int symbol;
    
    for ( unsigned int modelId = 0;
          modelId < ptree->t_table->getNumberModels(); ++modelId ){
        int size = ptree->t_table->getSequencesLength(modelId) +
                   ptree->t_table->getInvariantsLength(modelId);
        const array2D<double> & table =
                ptree->t_model->getEquivalencyTable(modelId);
        for ( int site = 0; site < size; ++site ) {
            symbol = identical[modelId][site];
            for ( int state = 0;
                  state < ptree->t_model->getNumberStates(modelId); ++state ) {
                partialLikelihood[modelId]( site, state, 0 ) =
                        table( symbol, state );
            }
        } // end site
    } // end model
    calculated = true;
}



bool Node::makeParent( Node * newParent, double dist ) {
    Node * oldParent = parent;
    double oldParentDistance = pdistance;
    int j;

    // look for the child which will be the new parent.
    // false is returned if no child is found equal to the newParent
    for ( int i = 0; i < 3; ++i ) {
        // if the new parent is this child
        // or if the node is to become the new root (newParent == NULL)
        // and is not the actual root
        if ( child[i] == newParent ) {
            //the old parent becomes one of the child
            child[i] = oldParent;
            cdistance[i] = oldParentDistance;
            //install the new parent
            parent = newParent;
            pdistance = dist;
            calculated = false;
            //if this node wasn't the root node (oldParent == NULL)
            if ( oldParent != NULL ) {
                //then for the oldParent (which is now child[i])
                //it is a recursive call which will end at the root.
                child[i]->makeParent( this, oldParentDistance );
                if ( ptree->t_table != NULL ) {
                    //thanks to the recursive call, the identical of the new
                    //children is computed before
                    computeIdentical();
                }
                return true;
            }
            //if it was the root node
            else {
                for ( j = i; j < 2; ++j ) {
                    child[j] = child[j + 1];
                    cdistance[j] = cdistance[j + 1];
                }
                child[2] = NULL;
                cdistance[2] = 0.0;
                if ( ptree->t_table != NULL ) {
                    //the identical of the two children left was up to date
                    //but there is now one child less. (two instead of three)
                    computeIdentical();
                }
                return true;
            } // end of the case were child[i] was the root (end of recursion)
        } // end of if child[i] == new parent
    } // end of looking for the child which is the new parent
    assert( 0 );
    return ( false );
}



void Node::optimizedLikelihoodLeafParent() {
    // Optimization not applicable
    assert ( ( child[2] == NULL ) && ( child[0]->isLeaf() ) &&
    ( child[1]->isLeaf() ) && !calculated );

    int left, right;
    double temp1, temp2;
    
    int numberModels = ptree->t_table->getNumberModels();

    array2D < int > fastLookup;

    array3D < double > * look1, * look2;

    // Initialize lookup and fastLookup table
    for ( int modelId = 0; modelId < numberModels; ++modelId ) {
        int numberRates = ptree->t_model->getNumberRateCategories( modelId );
        int numberStates =  ptree->t_model->getNumberStates( modelId );
        for ( int rate = 0; rate < numberRates; ++rate ) {
            for ( int initialState = 0; initialState < numberStates;
                  ++initialState ) {
                for ( int finalState = 0; finalState < numberStates;
                      ++finalState ) {
                    lookup[0][modelId]( rate, initialState, finalState ) =
                        ptree->t_model->probability( initialState, finalState,
                                                 cdistance[0], rate, modelId );
                    lookup[1][modelId]( rate, initialState, finalState ) =
                        ptree->t_model->probability( initialState, finalState,
                                                 cdistance[1], rate, modelId );
                }
            }
        }
        //fastLookuptable
        int numberSymbols = ptree->t_model->getNumberSymbols( modelId );
        fastLookup.resize( numberSymbols, numberSymbols );
        for ( left = 0; left < numberSymbols; ++left ) {
            for ( right = 0; right < numberSymbols; ++right ) {
                fastLookup( left, right ) = -1;
            }
        }
        
        int sequencesLength = ptree->t_table->getSequencesLength(modelId);
        int invariantsLength = ptree->t_table->getInvariantsLength(modelId);    
        look1 = & lookup[0][modelId];
        look2 = & lookup[1][modelId];

        for ( int site = 0; site < sequencesLength + invariantsLength; ++site ){
            left = child[0]->identical[modelId][site];
            right = child[1]->identical[modelId][site];
            if ( fastLookup( left, right ) == -1 ) {
                for ( int rate = 0; rate < numberRates; ++rate ) {
                    for ( int initialState = 0; initialState < numberStates;
                          ++initialState ) {
                        temp1 = 0.0;
                        temp2 = 0.0;
                        for ( int finalState = 0; finalState < numberStates;
                              ++finalState ) {
                            temp1 += (( * look1 )(rate, initialState, finalState) *
                                 child[0]->partialLikelihood[modelId]
                                 ( site, finalState, 0 ) );
                            temp2 += (( * look2 )(rate, initialState, finalState)*
                                 child[1]->partialLikelihood[modelId]
                                 ( site, finalState, 0 ) );
                        }
                        partialLikelihood[modelId]( site, initialState, rate ) =
                           temp1 * temp2;
                    } // for each initial state
                } // for each rate
                fastLookup( left, right ) = site;
            }
            else {
                memcpy(&(partialLikelihood[modelId](site,0,0)),
                   &(partialLikelihood[modelId](fastLookup( left, right ),0,0)),
                   sizeof(double)*numberRates*numberStates);
#ifdef DEBUG
                for ( int i = 0; i < numberStates; ++i ){
                    for ( int j = 0; j < numberRates; ++j ){
                        assert(partialLikelihood[modelId](site,i,j) ==
                    partialLikelihood[modelId](fastLookup( left, right ),i,j));
                    }
                }
#endif
            }   //if in fastlookup
#ifdef DEBUG
            double test = 0.0;
            for ( int rate = 0; rate < numberRates; ++rate ) {
                for ( int initialState = 0; initialState < numberStates;
                          ++initialState ) {
                    test +=
                        partialLikelihood[modelId]( site, initialState, rate );
                }
            }
            if (test==0.0){
                cerr << "site:" << site << endl; 
                for ( int finalState = 0; finalState < numberStates;
                      ++finalState ) {
                    cerr << child[0]->partialLikelihood[modelId]( site, finalState, 0 ) << endl;
                }
                for ( int finalState = 0; finalState < numberStates;
                      ++finalState ) {
                    cerr << child[1]->partialLikelihood[modelId]( site, finalState, 0 ) << endl;
                }
                for ( int initialState = 0; initialState < numberStates;
                      ++initialState ) {
                    cerr << "initialState = " << initialState << endl;
                    cerr << ( * look1 )(4, initialState, 0) << endl;
                    cerr << ( * look2 )(4, initialState, 2) << endl;
                }
                assert(0);
            }
#endif
        } // for each site
    } // for each model
    calculated = true;
}

bool Node::matches_clade( vector < int > & clade_rep, string & sclade_rep ) {
    assert( 0 );
    clade_rep = clade_rep;
    sclade_rep = sclade_rep;

    /*
    if ( clade_rep.length() != clade.length() ) {
    cerr << "Error in clade representation" << endl;
    exit( 1 );
    }
    if ( clade_rep == clade ) {
    sclade_rep = label;
    return true;
    }
    */

    return false;
}

void Node::build_clade( int mode, SequenceTable & data ) {
    assert( 0 );
    mode = mode;
    data = data;

    /*
    int i;
    unsigned int j;
    char num[10];

    clade.resize( data.numberSpecies );

    if ( isLeaf() ) {
    for ( i = 0; i < data.numberSpecies; i++ ) {
    clade[i] = 0;
    }
    if ( mode == 0 ) {
    //      label = toStringTopology() ;
    for ( j = 0; j < label.length(); j++ ) num[j] = label[j];
    num[j] = 0;
    clade[atoi( num ) - 1] = 1;
    }
    else {
    clade[cnumber - 1] = 1;
    }
    return;
    }

    if ( child[2] != NULL ) {
    for ( i = 0; i < data.numberSpecies; i++ ) {
    if ( child[0]->clade[i] || child[1]->clade[i] ||
    child[2]->clade[i] ) clade[i] = 1;
    else
    clade[i] = 0;
    }
    return;
    }
    for ( i = 0; i < data.numberSpecies; i++ ) {
    if ( child[0]->clade[i] || child[1]->clade[i] ) {
    clade[i] = 1;
    }
    else {
    clade[i] = 0;
    }
    }
    */
}





