#include "Tree/SimulationTree.h"

#include "Tree/Node.h"

#include "Models/Model.h"

SimulationTree::SimulationTree() : Tree() {
}

SimulationTree::SimulationTree( string stree ) : Tree( stree ) {
}

bool SimulationTree::simulate( const vector<int> length ) {

    assert( t_model );

    Singleton < randombox > & randBox = Singleton < randombox >::instance();
    
    unsigned int numberModels = length.size();    
    vector< vector<string> > rootSequence(numberModels);

    categories.resize(numberModels);

    for ( unsigned int modelId = 0; modelId < numberModels; ++modelId ) {
        for ( int site = 0; site < length[modelId]; ++site ) {
            /* assign a rate category */
            int rate = 0;
            double number = randBox.ran() -
                    t_model->getRateCategoryProbability( rate, modelId );
            while ( number > 0 ) {
                ++rate;
                number -= t_model->getRateCategoryProbability(rate, modelId);
            }
            assert( rate < t_model->getNumberRateCategories(modelId) );
            categories[modelId].push_back(rate);
        }
        vector< vector<int> > rubbish;
        result.clear();
        root->simulate( categories, rubbish, result );
    }
    return true;
}

void SimulationTree::printSequenceTable( ostream & out ) {
/*
    int seqLength = 0;
    for ( unsigned int modelId = 0; modelId < result[0].size(); ++modelId ) {
        for ( unsigned int site = 0; site< result[i].size(); ++site ){
            seqLength += t_model->getSymbol(result[0][modelId][site],modelId);
        }
    }
*/
    out << endl;
    for ( unsigned int modelId = 0; modelId < categories.size(); ++modelId ) {
        out << endl << "#";
        for ( unsigned int site = 0; site< categories[modelId].size(); ++site ){
            out << categories[modelId][site];
        }        
    }

    out << endl << endl;
    for (int i = 0; i < getNumberLeaves(); ++i) {
        out << leafRefVector[i]->label << endl;
        for (unsigned int modelId = 0; modelId < result[0].size(); ++modelId){
            for (unsigned int site = 0; site< result[i][modelId].size(); ++site){
                out << t_model->getSymbol(result[i][modelId][site],modelId);
            }
            out << endl;
        }
        out << endl;
    }
}

void SimulationTree::loadModel( Model * pmodel ) {
    assert ( t_model == NULL );
    t_model = pmodel;
}



