#include "Tree/Tree.h"

#include <assert.h>

#include <math.h>


#include <algorithm>
#include <functional>
#include <iostream>

/*tolower function*/
#include <cctype>
#include <cstdlib>

using namespace std;


#include "Models/Model.h"

#include "Tree/depthmap.h"

#include "Tree/Node.h"

double Tree::loglikelihood() {
    assert ( t_table && t_model );

    unsigned int i;

    //BEWARE, the order of processing is important
    //leafRefVector and nodeRefVector are made in order to
    //compute the children likelihood before
    for ( i = 0; i < leafRefVector.size(); ++i ) {
#ifdef DEBUG
        assert(leafRefVector[i]->index == i);
#endif
        if ( !( leafRefVector[i]->calculated ) ) {
            leafRefVector[i]->likelihood();
        }
    }
    for ( i = 0; i < nodeRefVector.size(); ++i ) {
#ifdef DEBUG
        assert(nodeRefVector[i]->index == i);
#endif
        if ( !( nodeRefVector[i]->calculated ) ) {
            nodeRefVector[i]->likelihood();
        }
    }


    double likelihood = 0.0;
    for( unsigned int modelId = 0; modelId < t_table->getNumberModels(); ++modelId ){
        int sequenceLength = t_table->getSequencesLength(modelId);
        int numberRateCategories = t_model->getNumberRateCategories(modelId);
        double rateProb[numberRateCategories];                
        for ( int rateCat = 0; rateCat < numberRateCategories; ++rateCat ) {
            rateProb[rateCat] =
            t_model->getRateCategoryProbability( rateCat, modelId );
        }
        int numberStates = t_model->getNumberStates(modelId);
        double frequency[numberStates][numberRateCategories];
        for ( int state = 0; state < numberStates; ++state ) {
            for ( int rateCat = 0; rateCat < numberRateCategories; ++rateCat ) {
                frequency[state] [rateCat] = t_model->getFrequency( state,
                                                         rateCat, modelId );
            }
        }

        for ( int site = 0; site < sequenceLength; ++site ) {
            double tempr = 0.0;
            for ( int rate = 0; rate < numberRateCategories; ++rate ) {
                double tempn = 0.0;
                for ( int nucleotide = 0; nucleotide < numberStates; ++nucleotide ) {
                    tempn += ( (frequency[nucleotide][rate])*
                root->partialLikelihood[modelId](site, nucleotide, rate ) );
                }
                tempr += ( rateProb[rate] * tempn );
            } // end rate
            likelihood += log( tempr );
        }//end site
        

        //invariant
        const vector< pair< string, unsigned int > > &inv =
            t_table->getInvariantBases(modelId);
        for ( unsigned int invId = 0; invId < inv.size(); ++invId ) {
            double tempr = 0.0;
            for ( int rate = 0; rate < numberRateCategories; ++rate ) {
                double tempn = 0.0;
                for ( int nucleotide = 0; nucleotide < numberStates;
                      ++nucleotide ) {
                    tempn += ( frequency[nucleotide] [rate]  *
                    root->partialLikelihood[modelId]( sequenceLength+invId,
                                                         nucleotide, rate ) );
                }
                tempr += ( rateProb[rate] * tempn );
            }
            likelihood += ( log( tempr ) * inv[invId].second );
        } // end invId
    }

    return ( likelihood );
}

double Tree::loglikelihood( vector < double > & bl ) {
    double ans;
    vector < double > old;
    getBranchVector( old );
    setBranchVector( bl );
    ans = loglikelihood();
    setBranchVector( old );
    return ( ans );
}

Tree::Tree() {
    root = NULL;
    t_model = NULL;
    t_table = NULL;
}

Tree::Tree( string stree ) {
    root = NULL;
    t_model = NULL;
    t_table = NULL;

    // Create a depth map of the tree
    int i, n = 0;
    depthmap map( stree );

    // Create the nodes
    root = new Node( this );

    for ( i = 1; i < map.length; i++ ) {
        if ( map.depths[0] == ( map.depths[i] - 1 ) ) {
            root->child[n] = Node::create_node( map, i, this );
            root->child[n]->parent = root;
            root->child[n]->pdistance = root->cdistance[n] = map.distance[i];
            n++;
        }
    }
    createIndex();
}

Tree::Tree(  const Tree & src ) {
    assert( 0 );
    *this=src;
}

Tree::~Tree(){
    if (root){
        delete root;
    }
    t_model = NULL;
    t_table = NULL;
}

bool Tree::greater( string & ostring1, string & ostring2,
string & soutg ) {

    string string1;
    string string2;
    string outg;

    string1.resize( ostring1.length() );
    string2.resize( ostring2.length() );
    outg.resize( soutg.length() );

    int bracketCount1 = 0;
    int bracketCount2 = 0;

    transform<string::iterator,string::iterator,int(*)(int)>
        ( ostring1.begin(), ostring1.end(), string1.begin(), tolower );
    transform<string::iterator,string::iterator,int(*)(int)>
        ( ostring2.begin(), ostring2.end(), string2.begin(), tolower );
    transform<string::iterator,string::iterator,int(*)(int)>
        ( soutg.begin(), soutg.end(), outg.begin(), tolower );

    if ( string1 == outg ) return false;
    if ( string2 == outg ) return true;

    bracketCount1 = count
        ( string1.begin(), string1.end(), '(' );
    bracketCount2 = count
        ( string2.begin(), string2.end(), '(' );

    if ( bracketCount1 > bracketCount2 ) {
        return true;
    }
    if ( bracketCount1 < bracketCount2 ) {
        return false;
    }
    return ( string1 > string2 );
}

bool Tree::standardizeTree( string newOutgroup ) {

    string left, mid, right;

    Node * tempn;
    double temp;

    installOutgroup( newOutgroup );
    
    for ( unsigned int i = 0; i < nodeRefVector.size() - 1; ++i ) {
        left = nodeRefVector[i]->child[0]->toStringTopology();
        right = nodeRefVector[i]->child[1]->toStringTopology();

        // Swap branches around if necessary
        if ( greater( left, right, newOutgroup ) ) {
            tempn = nodeRefVector[i]->child[0];
            nodeRefVector[i]->child[0] = nodeRefVector[i]->child[1];
            nodeRefVector[i]->child[1] = tempn;

            temp = nodeRefVector[i]->cdistance[0];
            nodeRefVector[i]->cdistance[0] = nodeRefVector[i]->cdistance[1];
            nodeRefVector[i]->cdistance[1] = temp;
        }
    }

    left = root->child[0]->toStringTopology();
    mid = root->child[1]->toStringTopology();


    if ( greater( left, mid, newOutgroup ) ) {
        tempn = root->child[0];
        root->child[0] = root->child[1];
        root->child[1] = tempn;
        temp = root->cdistance[0];
        root->cdistance[0] = root->cdistance[1];
        root->cdistance[1] = temp;
    }


    left = root->child[0]->toStringTopology();
    right = root->child[2]->toStringTopology();

    if ( greater( left, right, newOutgroup ) ) {
        tempn = root->child[0];
        root->child[0] = root->child[2];
        root->child[2] = tempn;
        temp = root->cdistance[0];
        root->cdistance[0] = root->cdistance[2];
        root->cdistance[2] = temp;
    }

    mid = root->child[1]->toStringTopology();
    right = root->child[2]->toStringTopology();

    if ( greater( mid, right, newOutgroup ) ) {
        tempn = root->child[1];
        root->child[1] = root->child[2];
        root->child[2] = tempn;
        temp = root->cdistance[1];
        root->cdistance[1] = root->cdistance[2];
        root->cdistance[2] = temp;
    }

    createIndex();
    return true;
}

void Tree::addLeaf( const string & leafLabel, int branchId,
double distParent, double fracBranch ) {

    assert( ( fracBranch >= 0.0 ) && ( fracBranch <= 1.0 ) );

    Node * newLeaf = new Node( leafLabel, this );
    Node * internalNode; // the new node usually required to add a leaf node

    int numberBranches = getNumberBranches();

    int nodeId;
    int side;

    //create the root if necessary
    if ( root == NULL ) {
        root = new Node( this );
    }

    // if the root has less than 3 children (construction case)
    for ( int i = 0; i < 3; i++ ) {
        if ( ( root->child[i] == NULL ) && ( branchId == i ) ) {
            root->child[i] = newLeaf;
            newLeaf->parent = root;
            root->cdistance[i] = newLeaf->pdistance = distParent;
            if ( ( root->child[0] != NULL ) && ( root->child[1] != NULL ) &&
            ( root->child[2] != NULL ) ) {
                createIndex();
                if ( t_table != NULL ){
                    root->computeIdentical();
                }
            }
            return;
        }
    }

    //if the root has more than three child
    internalNode = new Node( this ); // the new node

    // Error
    assert ( ( branchId < numberBranches ) && ( branchId >= 0 ) );

    // find the branch in the tree (the parent and the child linked by the
    // branch)
    retrieveBranchEnds( branchId, nodeId, side );

    double distance = nodeRefVector[nodeId]->cdistance[side];

    internalNode->parent = nodeRefVector[nodeId];

    internalNode->pdistance = distance * fracBranch;
    nodeRefVector[nodeId]->cdistance[side] = internalNode->pdistance;

    internalNode->child[0] = nodeRefVector[nodeId]->child[side];
    internalNode->child[0]->parent = internalNode;
    internalNode->cdistance[0] = distance * ( 1.0 - fracBranch );
    internalNode->child[0]->pdistance = internalNode->cdistance[0];

    nodeRefVector[nodeId]->child[side] = internalNode;

    internalNode->child[1] = newLeaf;
    internalNode->child[1]->parent = internalNode;
    internalNode->cdistance[1] = distParent;
    internalNode->child[1]->pdistance = distParent;

    internalNode->child[2] = NULL;
    internalNode->cdistance[2] = 0.0;

    newLeaf->calculated = false;
    Node * temp = newLeaf->parent;
    while ( temp != NULL ) {
        temp->calculated = false;
        if ( t_table != NULL ) {
            temp->computeIdentical();
        }
        temp = temp->parent;
    }
    // Rebuild index
    createIndex();
}

bool Tree::removeLeaf( const string & leafLabel ) {

    int j;

    Node * leafNode = NULL;
    int leafChildId;

    Node * leafParent;
    int parentChildId;

    Node * leafGrandParent = NULL;

    unsigned int leafId = 0;

    while ( ( leafId < leafRefVector.size() ) && ( !leafNode ) ) {
        if ( leafRefVector[leafId]->label == leafLabel ) {
            leafNode = leafRefVector[leafId];
        }
        ++leafId;
    }

    if ( !leafNode ) {
        cerr << "Unable to find the leaf " << leafLabel << endl;
        exit( 1 );
    }

    leafParent = leafNode->parent;

    assert( leafParent != NULL );

    leafChildId = -1;
    for ( j = 0; j < 3; j++ ) {
        if ( leafParent->child[j] == leafNode ) {
            leafChildId = j;
        }
    }

    assert( leafChildId != -1 );

    //if the parent is the root
    if ( leafParent == root ) {
        assert( nodeRefVector.size() != 1 );
        //reroot the tree to another node
        nodeRefVector[nodeRefVector.size() - 2]->makeRoot();
        root = nodeRefVector[nodeRefVector.size() - 2];
        createIndex();
        //and try again to remove the leaf
        return ( removeLeaf( leafLabel ) );
    }
    //otherwise
    else {
        leafGrandParent = leafParent->parent;
        assert( leafGrandParent != NULL );

        parentChildId = -1;

        for ( j = 0; j < 3; j++ ) {
            if ( leafGrandParent->child[j] == leafParent ) {
                parentChildId = j;
            }
        }

        assert( parentChildId != -1 );

        //to remove a leaf : destruct the leaf, the second child of the
        //leaf's parent becomes the child of the leaf's grandparent instead
        //of the leaf's parent
        leafGrandParent->child[parentChildId] =
        leafParent->child[( leafChildId == 0 ? 1 : 0 )];
        leafGrandParent->cdistance[parentChildId] +=
        leafParent->cdistance[( leafChildId == 0 ? 1 : 0 )];
        leafGrandParent->child[parentChildId]->parent = leafGrandParent;
        leafGrandParent->child[parentChildId]->pdistance =
        leafGrandParent->cdistance[parentChildId];

        //prevent unwanted destruction/modification when destroying the leaf
        //and its parent
        leafParent->parent = NULL;
        leafParent->child[( leafChildId == 0 ? 1 : 0 )] = NULL;
        //this call will destruct the parent and the leaf
        delete leafParent;


        //update the status of the ancestors
        Node * temp = leafGrandParent;
        while ( temp != NULL ) {
            temp->calculated = false;
            if ( t_table != NULL ) {
                temp->computeIdentical();
            } // if t_table != NULL
            temp = temp->parent;
        }
        createIndex();
        return true;
    }
    assert( 0 );
    return false;
}


bool Tree::addClusteredNode( Node * node, int branchId, double distParent,
                             double fracBranch ) {

    assert(fracBranch>=0.0);
    assert(fracBranch<=1.0);

    Node * internalNode = NULL;

    if ( root == NULL ){
        root = new Node( this );
    }

    node->clustered = true;

    //First three nodes
    for ( int i = 0; i < 3; ++i ) {
        if ( ( root->child[i] == NULL ) && ( branchId == i ) ) {
            root->child[i] = node;
            node->parent = root;
            root->cdistance[i] = node->pdistance = distParent;
            if ( ( root->child[0] != NULL ) &&
                 ( root->child[1] != NULL ) &&
                 ( root->child[2] != NULL ) ) {
                createIndex();
                root->computeIdentical();
            }
            return true;
        }
    }

    // Error
    if ( (branchId >= getNumberBranches()) || (branchId < 0) ) {
        cerr << "Error: Attempt to add a node to branch index "<< branchId << endl;
        cerr << "Number of branches = " << getNumberBranches() << endl;
        exit( 1 );
    }

    //addition near the root
    int nodeId;
    int sonId;
    
    if ( branchId >= ( getNumberBranches() - 3 ) ) {
       nodeId = getNumberNodes() - 1;
       sonId = branchId - getNumberBranches() + 3;
    }
    else {
       nodeId = branchId / 2;
       sonId = ( branchId % 2 );
    }
    //prevent addition in another cluster
    Node* ancestor = nodeRefVector[nodeId];
    while(ancestor){
        if ( ancestor->clustered == true ){
            return false;
        }
        ancestor = ancestor->parent;
    }
    
    //if the root has more than three child
    internalNode = new Node( this ); // the new node
    
    double distance = nodeRefVector[nodeId]->cdistance[sonId];

    //add the internal node: split branchId and and the new node to internalNode
    internalNode->parent = nodeRefVector[nodeId];
    nodeRefVector[nodeId]->cdistance[sonId] = distance * fracBranch;
    internalNode->pdistance = nodeRefVector[nodeId]->cdistance[sonId];

    internalNode->child[0] = nodeRefVector[nodeId]->child[sonId];
    internalNode->child[0]->parent = internalNode;
    internalNode->cdistance[0] = distance * (1.0-fracBranch);
    internalNode->child[0]->pdistance = internalNode->cdistance[0];

    nodeRefVector[nodeId]->child[sonId] = internalNode;

    internalNode->child[1] = node;
    node->parent = internalNode;
    internalNode->cdistance[1] = distParent;
    node->pdistance = distParent;


    //invalidate likelihood and compute identical
    node->calculated = false;
    Node * temp = node->parent;
    while ( temp != NULL ){
        temp->calculated = false;
        if ( t_table != NULL ) {
            temp->computeIdentical();
        }
        temp = temp->parent;
    }
    // Rebuild index
    createIndex();
    return true;
}

Node * Tree::removeNode( const string & stringTopology ) {

    Node * ret = NULL;
    for( unsigned int i = 0; i < leafRefVector.size(); ++i ){
        if( stringTopology == leafRefVector[i]->toStringTopology() ){
            ret = leafRefVector[i];
            break;
        }
    }
    for( unsigned int i = 0; i < nodeRefVector.size(); ++i ){
        if( stringTopology == nodeRefVector[i]->toStringTopology() ){
            ret = nodeRefVector[i];
            break;
        }
    }
    
    if ( !ret ) {
        cerr << "Unable to find the cluster " << stringTopology << endl;
        exit( 1 );
    }
    
    Node* clusterParent = ret->parent;
    assert ( clusterParent );
    int clusterChildId = -1;
    for ( int childId = 0; childId < 3; ++childId ) {
        if ( clusterParent->child[childId] == ret ) {
            clusterChildId = childId;
        }
    }
    assert( clusterChildId != -1 );

    
    //if the parent is the root
    if ( clusterParent == root ) {
        assert(0);
        assert( nodeRefVector.size() != 1 );
        //reroot the tree to another node
        nodeRefVector[nodeRefVector.size() - 2]->makeRoot();
        root = nodeRefVector[nodeRefVector.size() - 2];
        createIndex();
        //and try again to remove the cluster
        return ( removeNode( stringTopology ) );
    }
    else{
        Node* clusterGrandParent = clusterParent->parent;
        int parentChildId = -1;
        for ( int childId = 0; childId < 3; ++childId ) {
            if ( clusterGrandParent->child[childId] == clusterParent ) {
                parentChildId = childId;
            }
        }
        assert( parentChildId != -1 );
        
        //to remove a subtree : destruct the subtree, the second child of the
        //ancestor's parent becomes the child of the ancestor's grandparent
        //instead of the ancestor's parent
        //                *                             *
        //              /   \                         /   \
        //                   *            ----->            
        //                 /   \                             \
        //                *     *                             *
        clusterGrandParent->child[parentChildId] =
            clusterParent->child[( clusterChildId == 0 ? 1 : 0 )];
        clusterGrandParent->cdistance[parentChildId] +=
        clusterParent->cdistance[( clusterChildId == 0 ? 1 : 0 )];
        clusterGrandParent->child[parentChildId]->parent = clusterGrandParent;
        clusterGrandParent->child[parentChildId]->pdistance =
        clusterGrandParent->cdistance[parentChildId];
        
        //prevent unwanted destruction/modification when destroying the internal
        //node
        clusterParent->parent = NULL;
        clusterParent->child[( clusterChildId == 0 ? 1 : 0 )] = NULL;
        //we do not want to destroy the removed node
        clusterParent->child[clusterChildId] = NULL;
        //this call will destruct the parent only
        delete clusterParent;
        
        //update the status of the ancestors
        Node * temp = clusterGrandParent;
        while ( temp != NULL ) {
            temp->calculated = false;
            if ( t_table != NULL ) {
                temp->computeIdentical();
            } // if t_table != NULL
            temp = temp->parent;
        }
        createIndex();
        //return the removed node
        return ret;
    }
    assert( 0 );
    return false;
}

bool Tree::is_internal_branch( int branch ) {
    int branchId = branch % getNumberBranches();

    int nodeId;
    int side;

    // retrieve the node where the branch starts (nodeId)
    // and the child where it stops (side)

    // if branchId is a branch linking to the root
    if ( branchId >= ( getNumberBranches() - 3 ) ) {
        nodeId = getNumberNodes() - 1;
        side = branchId - getNumberBranches() + 3;
    }
    // otherwise
    else {
        nodeId = branchId / 2;
        side = ( branchId % 2 );
    }
    return ( !( nodeRefVector[nodeId]->child[side]->isLeaf() ) );
}

bool Tree::swapBranches( int branchId1, int branchId2 ) {

    int fromNode1, childNumber1, fromNode2, childNumber2;

    int numberBranches = getNumberBranches();

    Node * temp;
    double tempLength;

    bool ancestor1 = false;
    bool ancestor2 = false;

    assert( branchId1 >= 0 );
    assert( branchId2 >= 0 );
    assert( branchId1 < numberBranches );
    assert( branchId2 < numberBranches );

    retrieveBranchEnds( branchId1, fromNode1, childNumber1 );
    retrieveBranchEnds( branchId2, fromNode2, childNumber2 );

    // Swapping neighbouring branches causes no real change
    if ( ( fromNode1 == fromNode2 ) ||
    ( nodeRefVector[fromNode1]->child[childNumber1] == nodeRefVector[fromNode2] ) ||
    ( nodeRefVector[fromNode2]->child[childNumber2] ==
    nodeRefVector[fromNode1] ) ) {
        return ( false );
    }


    // Check for ancestor situation
    temp = nodeRefVector[fromNode1]->parent;
    while ( ( temp != NULL ) && ( !ancestor2 ) ) {
        if ( temp == nodeRefVector[fromNode2]->child[childNumber2] ) {
            ancestor2 = true;
        }
        temp = temp->parent;
    }
    if ( ! ancestor2 ) {
        temp = nodeRefVector[fromNode2]->parent;
        while ( ( temp != NULL ) && ( !ancestor1 ) ) {
            if ( temp == nodeRefVector[fromNode1]->child[childNumber1] ) {
                ancestor1 = true;
            }
        }
        temp = temp->parent;
    }

    if ( ancestor1 || ancestor2 ) {
        int newBranchId1 = -1;
        int newBranchId2 = -1;

        if ( ancestor1 ) {
            Node * startNode = nodeRefVector[fromNode1];
            Node * endNode = nodeRefVector[fromNode1]->child[childNumber1];

            newBranchId2 = ( 2 * ( getNumberNodes() - 1 ) ) + childNumber2;

            nodeRefVector[fromNode2]->makeRoot();
            root = nodeRefVector[fromNode2];
            createIndex();

            assert( !endNode->isLeaf() );
            //if ( !endNode->isLeaf() ) {
            for ( int i = 0; i < 2; ++i ) {
                if ( endNode->child[i] == startNode )
                    newBranchId1 = ( 2 * endNode->index ) + i;
            }
            //}
            //else {
            //    for ( i = 0; i < 2; ++i ) {
            //        if ( startNode->child[i] == endNode )
            //            newBranchId2 = ( 2 * nstart->index ) + i;
            //    }
            //}


        }

        if ( ancestor2 ) {
            Node * startNode = nodeRefVector[fromNode2];
            Node * endNode = nodeRefVector[fromNode2]->child[childNumber2];

            newBranchId1 = ( 2 * ( getNumberNodes() - 1 ) ) + childNumber1;

            nodeRefVector[fromNode1]->makeRoot();
            root = nodeRefVector[fromNode1];
            createIndex();

            assert( !endNode->isLeaf() );
            for ( int i = 0; i < 2; ++i ) {
                if ( endNode->child[i] == startNode ) {
                    newBranchId2 = ( 2 * endNode->index ) + i;
                }
            }

            /*
            if ( !nend->isLeaf() ) {
            for ( i = 0; i < 2; i++ ) {
            if ( node[nend->index]->child[i] == nstart )
            nbranch2 = ( 2 * nend->index ) + i;
            }

            }
            else {
            for ( i = 0; i < 2; i++ ) {
            if ( node[nstart->index]->child[i] == nend )
            nbranch2 = ( 2 * nstart->index ) + i;
            }
            }
            */
        }
        return ( swapBranches( newBranchId1, newBranchId2 ) );
    }

    temp = nodeRefVector[fromNode1]->child[childNumber1];
    tempLength = nodeRefVector[fromNode1]->cdistance[childNumber1];
    nodeRefVector[fromNode1]->child[childNumber1] =
    nodeRefVector[fromNode2]->child[childNumber2];
    nodeRefVector[fromNode1]->cdistance[childNumber1] =
    nodeRefVector[fromNode2]->cdistance[childNumber2];
    nodeRefVector[fromNode1]->child[childNumber1]->parent = nodeRefVector[fromNode1];
    nodeRefVector[fromNode1]->child[childNumber1]->pdistance =
    nodeRefVector[fromNode1]->cdistance[childNumber1];

    nodeRefVector[fromNode2]->child[childNumber2] = temp;
    nodeRefVector[fromNode2]->cdistance[childNumber2] = tempLength;
    temp->parent = nodeRefVector[fromNode2];
    temp->pdistance = tempLength;

    temp = nodeRefVector[fromNode1];
    while ( temp != NULL ) {
        temp->calculated = false;
        temp = temp->parent;
    }
    temp = nodeRefVector[fromNode2];
    while ( temp != NULL ) {
        temp->calculated = false;
        temp = temp->parent;
    }


    createIndex();

    if ( t_table != NULL ) {
        for ( int start = 0; start < getNumberNodes(); ++start ) {
            nodeRefVector[start]->computeIdentical();
        }
    }
    return true;
}

bool Tree::isRooted() {
    return ( root->child[2] == NULL );
}

double Tree::insertBranch( int movedBranch, int destBranch,
double insertionFrac ) {
    int movedBranchFromNode, movedBranchSide;
    int destBranchFromNode, destBranchSide;

    int numberBranches = getNumberBranches();

    int i;

    //Find the nodes linked by the movedBranch
    if ( movedBranch >= ( numberBranches - 3 ) ) {
        movedBranchFromNode = getNumberNodes() - 1;
        movedBranchSide = movedBranch - numberBranches + 3;
    }
    else {
        movedBranchFromNode = movedBranch / 2;
        movedBranchSide = ( movedBranch % 2 );
    }

    //Find the nodes linked by the destBranch
    if ( destBranch >= ( numberBranches - 3 ) ) {
        destBranchFromNode = getNumberNodes() - 1;
        destBranchSide = destBranch - numberBranches + 3;
    }
    else {
        destBranchFromNode = destBranch / 2;
        destBranchSide = ( destBranch % 2 );
    }

    // Swapping neighbouring branches causes no real change
    if ( ( destBranchFromNode == movedBranchFromNode ) ||
    ( nodeRefVector[destBranchFromNode]->child[destBranchSide] ==
    nodeRefVector[movedBranchFromNode] ) ||
    ( nodeRefVector[movedBranchFromNode]->child[movedBranchSide] ==
    nodeRefVector[destBranchFromNode] ) ) {
        return ( -1.0 );
    }


    // if the moved branch is linking to the root or if the branch moved is an
    // ancestor of the dest branch, change the root first and then call the
    // function again with the new branches ID

    // Check for ancestor situation (is the branch moved an ancestor of
    // the dest branch ?)
    // if the movedBranch is not linked to the root node
    bool ancestor = false;
    if ( nodeRefVector[movedBranchFromNode] != root ) {
        Node * temp = nodeRefVector[destBranchFromNode]->parent;
        while ( ( temp != NULL ) && ( !ancestor ) ) {
            if ( temp == nodeRefVector[movedBranchFromNode]->child[movedBranchSide] ) {
                ancestor = true;
            }
            else {
                temp = temp->parent;
            }
        }
    }

    // if the moved branch is linking to the root swap the root and call
    // the function again
    if ( ( nodeRefVector[movedBranchFromNode] == root ) ||
    ( ancestor == true ) ) {
        Node * start = root;
        Node * end = root->child[movedBranchSide];

        //Change the root
        nodeRefVector[destBranchFromNode]->makeRoot();
        root = nodeRefVector[destBranchFromNode];
        createIndex();

        int newMovedBranch = -1;

        int newDestBranch = 2 * ( getNumberNodes() - 1 ) + destBranchSide;

        if ( !end->isLeaf() ) {
            for ( i = 0; i < 2; i++ ) {
                if ( nodeRefVector[end->index]->child[i] == start )
                    newMovedBranch = ( 2 * end->index ) + i;
            }
            for ( i = 0; i < 2; i++ ) {
                if ( nodeRefVector[start->index]->child[i] == end )
                    newMovedBranch = ( 2 * start->index ) + i;
            }
        }
        else {
            for ( i = 0; i < 2; i++ ) {
                if ( nodeRefVector[start->index]->child[i] == end )
                    newMovedBranch = ( 2 * start->index ) + i;
            }
        }

        assert( newMovedBranch < getNumberBranches() );
        assert( newDestBranch < getNumberBranches() );
        assert( newMovedBranch >= 0 );
        assert( newDestBranch >= 0 );

        return ( insertBranch( newMovedBranch, newDestBranch,
        insertionFrac ) );
    }


    // General Case : if the branch moved is NOT an ancestor of the dest branch
    // and is not linked to the root node
    Node * fromNode = nodeRefVector[movedBranchFromNode];
    Node * parent = fromNode->parent;

    int childBranch = -1;

    for ( i = 0; i < 3; i++ ) {
        if ( parent->child[i] == fromNode ) {
            childBranch = i;
        }
    }

    // Detach node and attach the other child to the parent
    int otherBranch = ( movedBranchSide == 0 ? 1 : 0 );
    parent->child[childBranch] = fromNode->child[otherBranch];
    parent->cdistance[childBranch] = parent->cdistance[childBranch] +
    fromNode->cdistance[otherBranch];
    parent->child[childBranch]->parent = parent;
    parent->child[childBranch]->pdistance = parent->cdistance[childBranch];


    fromNode->child[otherBranch] = NULL;
    fromNode->parent = NULL;

    // Find the MH (Metropolis-Hastings) Ratio : b/(x+y)
    // where b = length of arrival branch
    // x+y = length to parent + length to other child (from the node removed)
    double mh_ratio = nodeRefVector[destBranchFromNode]->cdistance[destBranchSide]
    / parent->cdistance[childBranch];

    // Reattach the node (split the arrival branch)
    double distance = nodeRefVector[destBranchFromNode]->cdistance[destBranchSide] *
    insertionFrac;

    //replace the removed child
    fromNode->child[otherBranch] =
    nodeRefVector[destBranchFromNode]->child[destBranchSide];
    fromNode->child[otherBranch]->parent = fromNode;
    fromNode->cdistance[otherBranch] =
    nodeRefVector[destBranchFromNode]->cdistance[destBranchSide] - distance;
    fromNode->child[otherBranch]->pdistance = fromNode->cdistance[otherBranch];

    //replace the parent
    nodeRefVector[destBranchFromNode]->child[destBranchSide] = fromNode;
    fromNode->parent = nodeRefVector[destBranchFromNode];
    nodeRefVector[destBranchFromNode]->cdistance[destBranchSide] = distance;
    fromNode->pdistance = distance;


    createIndex();

    if ( t_table != NULL ) {
        for ( int start = 0; start < getNumberNodes(); ++start ) {
            nodeRefVector[start]->calculated = false;
            nodeRefVector[start]->computeIdentical();
        }
    }

    // Return the non-negative MH ratio
    return mh_ratio;
}

bool Tree::unroot() {
    int id;
    if ( root->child[2] != NULL ) return false;
    if ( root->child[1]->isLeaf() ){
        if ( root->child[0]->isLeaf() ){
            assert(0);
            return false;
        }
        else{
            id = 0;
        }
    }
    else{
        id = 1;
    }
    Node * temp = root->child[id];
    double tempd = root->cdistance[id];
    root->child[id] = temp->child[0];
    root->child[2] = temp->child[1];
    root->child[id]->parent = root;
    root->child[2]->parent = root;
    root->cdistance[id] = root->child[id]->pdistance =
        temp->cdistance[0];
    root->cdistance[2] = root->child[2]->pdistance =
        temp->cdistance[1];
    root->cdistance[1-id] = root->child[1-id]->pdistance =
            root->cdistance[1-id] + tempd;
    temp->child[0] = temp->child[1] = temp->child[2] = NULL;
    temp->parent = NULL;
    delete temp;
    createIndex();

    return true;
}

double Tree::getBranchLength( int branchId ) const {
    int nodeId;
    int side;

    int numberBranches = getNumberBranches();

    if ( branchId >= ( numberBranches - 3 ) ) {
        nodeId = getNumberNodes() - 1;
        side = branchId - numberBranches + 3;
    }
    else {
        nodeId = branchId / 2;
        side = ( branchId % 2 );
    }
    return ( nodeRefVector[nodeId]->cdistance[side] );
}

void Tree::setBranchLength( int branchId, double newLength ) {

    int nodeId;
    int side;

    int numberBranches = getNumberBranches();

    // Find the node the branch is attached to
    if ( branchId >= ( numberBranches - 3 ) ) {
        nodeId = getNumberNodes() - 1;
        side = branchId - numberBranches + 3;
    }
    else {
        nodeId = branchId / 2;
        side = ( branchId % 2 );
    }

    nodeRefVector[nodeId]->cdistance[side] = newLength;
    nodeRefVector[nodeId]->child[side]->pdistance = newLength;

    Node * temp = nodeRefVector[nodeId];

    // Invalidate node likelihood calculations
    do {
        temp->calculated = false;
        temp = temp->parent;
    }
    while ( temp != NULL );
}

void Tree::getBranchVector( vector < double > &
branchesLength ) const {

    int nodeId;
    int side;

    int numberBranches = getNumberBranches();

    branchesLength.resize( numberBranches );

    for ( int i = 0; i < numberBranches; ++i ) {
        if ( i >= ( numberBranches - 3 ) ) {
            nodeId = getNumberNodes() - 1;
            side = i - numberBranches + 3;
        }
        else {
            nodeId = i / 2;
            side = ( i % 2 );
        }

        branchesLength[i] = nodeRefVector[nodeId]->cdistance[side];
    }
}

void Tree::setBranchVector( const vector < double > &
branchesLength ) {

    int nodeId;
    int side;

    int numberBranches = getNumberBranches();

    assert( (int)branchesLength.size() == numberBranches );

    // for each branch retrieve the father
    // and the child linked
    for ( int i = 0; i < numberBranches; i++ ) {
        if ( i >= ( numberBranches - 3 ) ) {
            nodeId = getNumberNodes() - 1;
            side = i - numberBranches + 3;
        }
        else {
            nodeId = i / 2;
            side = ( i % 2 );
        }

        nodeRefVector[nodeId]->cdistance[side] = branchesLength[i];
        nodeRefVector[nodeId]->child[side]->pdistance = branchesLength[i];
    }
    // all the nodes are to be computed again (but not the leaves)
    invalidateNodes();
}

string Tree::toStringClade( vector < int > * query_array, int size ) {
    return ( root->toStringClade( query_array, size ) );
}

string Tree::toString( bool rooted ) {
    if ( !rooted ) {
        return ( root->toString() );
    }
    else {
        char d1[64], d2[64], d3[64];
        sprintf( d1, "%0.4f", root->cdistance[0] / 2.0 );
        sprintf( d2, "%0.4f", root->cdistance[1] );
        sprintf( d3, "%0.4f", root->cdistance[2] );
        string ret = "(" + root->child[0]->toString() + ":" + d1 + ",(" +
        root->child[1]->toString() + ":" + d2 + "," +
        root->child[2]->toString() + ":" + d3 + "):" + d1 + ")";
        return ret;
    }
    //to avoid a warning
    return string();
}

string Tree::toStringTopology( bool rooted ) {
    if ( !rooted ) return ( root->toStringTopology() );

    string ret;
    ret = "(" + root->child[0]->toStringTopology() + ",(" +
    root->child[1]->toStringTopology() + "," +
    root->child[2]->toStringTopology() + "))";
    return ret;
}

string Tree::toStringNumbered( bool rooted ) {
    //  int i ;
    if ( rooted ) {
        return ( string( "(" + root->child[0]->toStringNumbered() + ",(" +
        root->child[1]->toStringNumbered() + "," +
        root->child[2]->toStringNumbered() + "))" ) );
    }
    else {
        return ( root->toStringNumbered() );
    }
}

void Tree::createIndex() {

    int nodeToExpandIndice;
    int i;
    unsigned int j;

    // construct the nodeRefVector and leafRefVector
    // ensure that for all (i,j) / i > j,
    // nodeRefVector[i] is NOT an ancestor of nodeRefVector[j]
    // this will help future computation when the state of an ancestor
    // requires the state of its childs to be known
    nodeRefVector.clear();
    leafRefVector.clear();

    if ( root != NULL ) {
        nodeRefVector.push_front( root );
    }
    else {
        return;
    }

    nodeToExpandIndice = 0;

    while ( nodeToExpandIndice >= 0 ) {
        // Expand out the node
        for ( i = 0; i < 3; i++ ) {
            if ( nodeRefVector[nodeToExpandIndice]->child[i] != NULL ) {
                if ( nodeRefVector[nodeToExpandIndice]->child[i]->isLeaf() ) {
                    leafRefVector.push_front( nodeRefVector[nodeToExpandIndice]->child[i] );
                }
                else {
                    //push the child at the front of the nodeRefVector dequeue
                    nodeRefVector.push_front( nodeRefVector[nodeToExpandIndice]->child[i] );
                    //update the nodeToExpandIndice value (there is a new node at the
                    //beginning of the dequeue)
                    ++nodeToExpandIndice;
                }
            }
        }
        //proceed through the next node
        --nodeToExpandIndice;
    }
    for ( j = 0 ; j < nodeRefVector.size() ; ++j ) {
        nodeRefVector[j]->index = j;
    }
    for ( j = 0 ; j < leafRefVector.size() ; ++j ) {
        leafRefVector[j]->index = j;
    }
}

void Tree::retrieveBranchEnds( int branchId, int & fromNode,
int & childNumber ) {
    int numberBranches = getNumberBranches();
    if ( branchId >= ( numberBranches - 3 ) ) {
        fromNode = getNumberNodes() - 1;
        childNumber = branchId - numberBranches + 3;
    }
    else {
        fromNode = branchId / 2;
        childNumber = ( branchId % 2 );
    }
}

bool Tree::installOutgroup( string newOutgroup ) {
    Node * leafNode = NULL;
    Node * leafParent = NULL;

    unsigned int leafId = 0;

    transform<string::iterator,string::iterator,int(*)(int)>
        ( newOutgroup.begin(), newOutgroup.end(), newOutgroup.begin(), tolower );

    while ( ( !leafParent ) && ( leafId < leafRefVector.size() ) ) {
        leafNode = leafRefVector[leafId];
        string leafLabel( leafNode->label );
        transform<string::iterator,string::iterator,int(*)(int)>
	    ( leafLabel.begin(), leafLabel.end(), leafLabel.begin(), tolower );
        if ( newOutgroup == leafLabel ) {
            leafParent = leafNode->parent;
        }
        ++leafId;
    }

    if ( !leafParent ) {
        cerr << "Error : No taxa match outgroup taxa \"" << newOutgroup << "\"" << endl;
        cerr << toString( false ) << endl;
        cerr << "Number of leaves = " << leafRefVector.size() << endl;
        for ( leafId = 0; leafId < leafRefVector.size(); ++leafId ) {
            cerr << ( leafId + 1 ) << "." << leafRefVector[leafId]->label << flush;
            if ( !leafRefVector[leafId]->isLeaf() ) {
                cerr << " is not a leaf for the tree : " << flush;
                cerr << leafRefVector[leafId]->toString();
            }
            cerr << endl;
        }
        exit( 1 );
    }

    this->outgroup = newOutgroup;

    if ( leafParent == root ) {
        return false;
    }
    else {          
        leafParent->makeRoot();
        root = leafParent;
        createIndex();    
        return true;
    }
    return 0;
}

void Tree::invalidateNodes() {
    for ( unsigned int i = 0; i < nodeRefVector.size(); ++i ) {
        nodeRefVector[i]->calculated = false;
    }
}

void Tree::invalidateLeaves() {
    for ( unsigned int i = 0; i < leafRefVector.size(); ++i ) {
        leafRefVector[i]->calculated = false;
    }
}


