#include "Tree/TreesSet.h"

#include "Tree/TreeModelValue.h"


TreesSet::TreesSet(int maxNumberEntries) :
vector<TreeModelValue *>(maxNumberEntries){
    assert(maxNumberEntries!=0);
    this->maxNumberEntries = maxNumberEntries;
    clear();
}

TreesSet::~TreesSet() {
    for ( iterator iter = begin(); iter != end(); ++iter ){
        delete (*iter) ;
    }
}


bool TreesSet::insert(const string& stringTree,
                      const vector <double>& branchLengths,
                      const vector <double>& modelParameters,
                      double logLikelihood) {
    //iterator used to store the position to insert the new element
    iterator insertIter;
    TreeModelValue* newEntry;
    //if the list is not empty find where to insert the element
    if ( size() != 0 ){
        insertIter = end();
        iterator compareIter = insertIter;
        //make compareIter point to the last element
        --compareIter;
        //in case the list is already full
        if ( (int)size() == maxNumberEntries ){
            // if the likelihood is lower than the lowest likelihood already
            // in then discard the new entry and exit
            if ( (*compareIter)->logLikelihood > logLikelihood ){
                return false;
            }
            //otherwise delete the last entry and reposition insertIter and
            //compareIter
            else{
                delete (*compareIter);
                pop_back();
                insertIter = end();
                compareIter = insertIter;
                --compareIter;
            }
        }
        //it will be a new entry so find where to put it
        while ( ( logLikelihood > (*compareIter)->logLikelihood ) &&
                ( insertIter != begin() ) ){
            insertIter = compareIter;
            if (compareIter != begin()){
                 --compareIter;
            }
        }
    }
    //if the list is empty
    else{
        //we will insert the new entry at the end
        insertIter = end();
    }
    newEntry = new TreeModelValue(stringTree, branchLengths,
                                             modelParameters, logLikelihood);
    vector <TreeModelValue *>::insert( insertIter, newEntry);
    return true ;
}


TreeModelValue * TreesSet::get(int index) {
    return (*this)[index-1];
}
