#include "depthmap.h"

using namespace std;

#include <string>

#include <iostream>

/** Start Depthmap Implementation *************** */

depthmap::depthmap(string tree) {
    int i = 0, depth = 0, j, node = 0;
    char clabel[32], number[32];
    int labelindex = 0;
    int * distance_stack, sp = 0;

    // Strip out junk
    while (tree[i] != '(') {
           i++;
    }
    number[0] = 0;

    // Assign as much space as will be needed for the depth.
    // This is actually certainly an overestimate but what the hell.

    length = tree.size() - i;

    depths = new int[length];
    identifier = new bool[length];
    label = new string[length];
    distance = new double[length];
    distance_stack = new int[length];

    // Initialise depth map
    for (j = 0; j < length; j++) {
        depths[j] = 0;
        identifier[j] = true;
        label[j] = string("");
        distance[j] = 0.0;
    }

    do {
        switch (tree[i]) {
            case '(':
                depths[node] = depth;
                identifier[node] = false; // Not a leaf node
                distance_stack[sp++] = node;
                depth++;
                node++;
                i++;
                break;
            case ')':
                depth--; // Closing bracket (adjust depth)
                distance[distance_stack[--sp]] = atof(number);
		number[0]=0;		
                i++;
                break;
            case ':':
                labelindex = 0;
                while (!isdigit(tree[i])) i++;

                while (isdigit(tree[i]) || tree[i] == '.') {
                    number[labelindex] = tree[i];
                    labelindex++;
                    i++;
                }
                number[labelindex] = 0;
            case ' ':
            case '\t':
            case '\n':
                while (isspace(tree[i])) i++;
                break;
            case ',':
                distance[distance_stack[--sp]] = atof(number);
		number[0]=0;		
                i++;
                break;
            default: // It's a leaf label
                labelindex = 0;
                while (!isspace(tree[i]) && (tree[i] != ':') && (tree[i] != ',') && (tree[i] != ')') && (i < length)) {
                    clabel[labelindex++] = tree[i];
                    i++;
                }
                clabel[labelindex] = 0;

                depths[node] = depth;
                identifier[node] = true; // Is a leaf node
                label[node] = string(clabel);
                distance_stack[sp++] = node;
                node++;
        }
    } while (depth > 0 && (unsigned int) i < tree.size());
    //  cout << "Made Map" << endl ;
    if (depth > 0) {
        cout << "Bad Tree : " << depth << endl;
        exit(1);
    }

    length = node;

    // Clean up
    delete[] distance_stack;
}

depthmap::~depthmap() {
    delete[] depths;
    delete[] label;
    delete[] identifier;
    delete[] distance;
}

void depthmap::printmap() {
    int i, index = 1, lindex = 1;
    for (i = 0; i < length; i++) {
        if (!identifier[i]) cout << "Node " << index++ << " Depth : " << depths[i] << " Distance : " << distance[i] << endl;
        else cout << "Leaf " << lindex++ << " Depth : " << depths[i] << " Label : " << label[i] << " Distance : " <<
               distance[i] << endl;
    }
    return;
}
