#include "Util/ParametersSet.h"


using namespace std;

#include <algorithm>
#include <functional>
#include <iostream>

ParametersSet::ParametersSet( const string & setName, bool autoRemoval ) :
name( setName ) {
    this->autoRemoval = autoRemoval;
}

int ParametersSet::intParameter( const string & parameter ) {
    char * endPtr;
    int result;
    iterator i = find( parameter );
    if ( i == end() ) {
        cerr << "missing parameter : \"" << parameter << "\" in the set "
        << name << endl;
        exit( 1 );
    }
    result = strtol( ( * i ).second.c_str(), & endPtr, 0 );
    if ( * endPtr != '\0' ) {
        cerr << "The parameter : \"" << parameter << "\" in the set "
        << name << " is not an integer" << endl;
        exit( 1 );
    }
    if ( autoRemoval ) {
        erase( i );
    }
    return result;
}

double ParametersSet::doubleParameter( const string & parameter ) {
    char * endPtr;
    double result;
    iterator i = find( parameter );
    if ( i == end() ) {
        cerr << "missing parameter : \"" << parameter << "\" in the set "
        << name << endl;
        exit( 1 );
    }
    result = strtod( ( * i ).second.c_str(), & endPtr );
    if ( * endPtr != '\0' ) {
        cerr << "The parameter : \"" << parameter << "\" in the set "
        << name << " is not an integer" << endl;
        exit( 1 );
    }
    if ( autoRemoval ) {
        erase( i );
    }
    return result;
}

string ParametersSet::stringParameter( const string & parameter ) {
    iterator i = find( parameter );
    if ( i == end() ) {
        cerr << "missing parameter : \"" << parameter << "\" in the set "
        << name << endl;
        exit( 1 );
    }
    string result = ( * i ).second;
    if ( autoRemoval ) {
        erase( i );
    }
    return result;
}

bool ParametersSet::boolParameter( const string & parameter ) {

    bool result;

    iterator i = find( parameter );
    if ( i == end() ) {
        cerr << "missing parameter : \"" << parameter << "\" in the set "
        << name << endl;
        exit( 1 );
    }
    string stringResult( ( * i ).second );
    for_each<string::iterator,int(*)(int)>( stringResult.begin(), stringResult.end(), tolower );

    if ( ( stringResult == "yes" )  || ( stringResult == "y" ) ||
    ( stringResult == "true" ) || ( stringResult == "t" ) ) {
        result = true;
    }
    else {
        if ( ( stringResult == "no" ) || ( stringResult == "n" ) ||
        ( stringResult == "false" ) || ( stringResult == "f" ) ) {
            result = false;
        }
        else {
            cerr << "The parameter : \"" << parameter << "\" in the set "
            << name << " can not be interpreted as a boolean,"
            << "hint:try yes/no" << endl;
            exit( 1 );
        }
    }
    if ( autoRemoval ) {
        erase( i );
    }
    return result;
}

void ParametersSet::addCategory( string categoryName ) {
    categories.push_back( ParametersSet( categoryName, autoRemoval ) );
}


bool ParametersSet::findCategory( string categoryName ){
    for ( vector < ParametersSet >::iterator i =
    categories.begin(); i != categories.end(); ++i ) {
        if ( ( * i ).name == categoryName ) {
            return true;
        }
    }
    return false;
}

ParametersSet & ParametersSet::operator () ( const string & categoryName ) {
    for ( vector < ParametersSet >::iterator i =
    categories.begin(); i != categories.end(); ++i ) {
        if ( ( * i ).name == categoryName ) {
            return ( * i );
        }
    }
    cerr << "category : \"" << categoryName << "\" not found in the set "
    << name << endl;
    exit( 1 );
    return ( * categories.end() );
}

const string & ParametersSet::getName() const {
    return name;
}

void ParametersSet::setName( const string& newName ) {
    name = newName;
}

void ParametersSet::saveToFile( ofstream & fileOutput, int indentation ) {
    string indent;
    indent.resize( indentation, ' ' );
    for (iterator i=begin(); i!=end(); ++i ){
        fileOutput << indent << (*i).first << " = " << (*i).second << endl;
    }
    for ( vector< ParametersSet >::iterator i=categories.begin();
          i!=categories.end(); ++i ){
        fileOutput << indent << "{" << (*i).getName() << "}" << endl;
        (*i).saveToFile( fileOutput, indentation+2 );
        fileOutput << indent << "{\\" << (*i).getName() << "}" << endl << endl;
    }
}


